"""
  Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import logging
import time
import datetime
import numpy as np
from enum import Enum

from wiliot_testers.wiliot_tester_tag_test import WiliotTesterTagTest
from wiliot_testers.wiliot_tester_tag_result import FailureCodes, WiliotTesterTagResultList
from wiliot_tools.test_equipment.test_equipment import Attenuator, YoctoSensor
from wiliot_core import InlayTypes
from wiliot_testers.offline.configs.global_vars_and_enums import TagTestingDefaults
from wiliot_testers.offline.modules.offline_r2r_controller import R2R


class TagTestingException(Exception):
    def __init__(self, msg):
        self.message = msg
        super().__init__(f'TagTesting Exception: {self.message}')


class R2RException(TagTestingException):
    def __init__(self, msg):
        super().__init__(f'R2R: {msg}')


class MissingLabelException(TagTestingException):
    def __init__(self, msg):
        super().__init__(f'Missing Label: {msg}')


class AttenuatorException(TagTestingException):
    def __init__(self, msg):
        super().__init__(f'Attenuator: {msg}')


class SensorException(TagTestingException):
    def __init__(self, msg):
        super().__init__(f'Sensor: {msg}')


MAX_MAIN_STATES = 5


class TagStates(Enum):
    IDLE = 0
    STOP = 99
    WAIT_MOH = 88
    GET_TRIGGER = 1
    START_TEST = 2
    TAG_TEST = 3
    PROCESS_TEST = 4
    END_TEST = 5
    WAIT_END = 6
    NEXT_TAG = 7


class TagTesting(object):
    def __init__(self, test_start_event, test_end_event, validation_end_event, stop_app_event, moh_object,
                 exception_queue, tag_results_queue, log_obj, test_object, user_config=None, hw_config=None):
        self.test_start_event = test_start_event
        self.test_end_event = test_end_event
        self.validation_end_event = validation_end_event
        self.stop_app_event = stop_app_event
        self.exception_q = exception_queue
        self.tag_results_q = tag_results_queue
        self.moh = moh_object
        self.test = test_object
        self.logger = logging.getLogger(log_obj.get_logger_name())
        self.user_config = user_config
        self.tested = 0
        self.missing_labels = 0
        self.missing_label_locations = []
        self.need_to_manual_trigger = False
        self.tag_test_running = False
        self.all_selected_tags = []
        self.all_duplicated_tags = []
        self.all_tags = []
        self.is_passed_list = []
        self.ttfp_list = []
        self.last_state = TagStates.IDLE
        self.is_gen_checked = False
        self.is_r2r = user_config['isR2R'].lower() == 'yes' and user_config['typeR2R'].lower() == 'gateway'

        self.is_init = False
        self.r2r = None
        try:
            # init tester class
            self.wiliot_tester = WiliotTesterTagTest(
                selected_test=self.user_config['testName'],
                logger_name=self.logger.name,
                logger_result_name=log_obj.get_results_logger_name(),
                logger_gw_name=log_obj.get_gw_logger_name(),
                stop_event_trig=self.stop_app_event,
                inlay=InlayTypes(self.user_config['inlay']),
                dir_for_gw_log=log_obj.get_test_folder(),
                gw_test_cmds=self.get_additional_gw_cmds(hw_config=hw_config),
            )
            self.gw_obj = self.wiliot_tester.get_gw_object()

            # init timeout for trigger:
            self.logger.info(f'Maximum time for gw trigger (i.e. missing label) '
                             f'is {TagTestingDefaults.TIMEOUT_FOR_MISSING_LABEL}')

            # init r2r if relevant:
            if self.is_r2r:
                self.r2r = R2R(exception_queue=exception_queue,
                               logger_name=log_obj.get_logger_name(),
                               exc_obj=R2RException,
                               gw_obj=self.gw_obj)

            # init external hw
            self.attenuator_handle = AttenuatorHandle(logger_name=self.logger.name, exception_q=exception_queue,
                                                      attenuator_configs=hw_config)
            self.sensors_handle = ExternalSensorHandle(logger_name=self.logger.name,
                                                       exception_q=exception_queue,
                                                       sensor_configs=hw_config)
            self.is_init = True

        except Exception as e:
            if isinstance(e, TagTestingException):
                self.exception_q.put(f'init: {e}', block=False)
            else:
                self.exception_q.put(TagTestingException(f'init: {e}').__str__(), block=False)

    @staticmethod
    def get_additional_gw_cmds(hw_config):
        if hw_config is None:
            return None
        if hw_config.get('printingBidirectionalMovement', 'No').lower() == 'yes':
            gpio = TagTestingDefaults.PRINTER_DIRECTION_GPIO.replace('.', '').upper()
            return [f'!cmd_gpio CONTROL_IN {gpio} 0']  # to get ack from printer cart

    def run(self):
        tester_res = WiliotTesterTagResultList()  # Empty list
        state = TagStates.IDLE
        if self.is_r2r:
            self.r2r.move_to_the_first_tag()

        while True:
            time.sleep(0)
            try:
                state = self.update_state(state)
                self.logger.info(f'TagTesting: start state {state}')
                if state == TagStates.GET_TRIGGER:
                    self.get_trigger()

                elif state == TagStates.START_TEST:
                    self.start_test()

                elif state == TagStates.TAG_TEST:
                    tester_res = self.tag_test()
                    self.need_to_manual_trigger = False

                elif state == TagStates.PROCESS_TEST:
                    self.process_test(tester_res=tester_res)

                elif state == TagStates.END_TEST:
                    self.end_test(tester_res=tester_res)

                elif state == TagStates.WAIT_END:
                    if self.is_r2r:
                        # wait till event end_of_validation is set
                        self.wait_for_new_event(
                            event=self.validation_end_event,
                            time_to_wait=TagTestingDefaults.EVENT_WAIT_TIME)
                        self.gw_obj.clear_rsp_str_input_q()

                elif state == TagStates.NEXT_TAG:
                    if self.is_r2r:
                        # send signal to the r2r to move
                        self.r2r.move_r2r()

                elif state == TagStates.STOP:
                    self.logger.info('Stop running TagTesting main loop')
                    break

                elif state == TagStates.WAIT_MOH:
                    self.wait_for_moh()

                self.logger.info(f'TagTesting: end state {state}')

            except Exception as e:
                if isinstance(e, TagTestingException):
                    self.exception_q.put(f'run: state{state.name}: {e}', block=False)
                else:
                    self.exception_q.put(TagTestingException(f'run: state{state.name}: {e}').__str__(), block=False)

                if state == TagStates.GET_TRIGGER:
                    self.need_to_manual_trigger = True
                elif state == TagStates.TAG_TEST:
                    tester_res = WiliotTesterTagResultList()
                    tester_res.set_total_fail_bin(fail_code=FailureCodes.GW_ERROR)
                    self.tag_test_running = False

                elif state == TagStates.PROCESS_TEST:
                    tester_res.set_total_fail_bin(fail_code=FailureCodes.SOFTWARE_GENERAL_ERROR, overwrite=True)
                elif state == TagStates.END_TEST or state == TagStates.WAIT_MOH:
                    self.logger.warning(f'TagTesting Got exception during {state}: {e}')
                elif state == TagStates.NEXT_TAG:
                    state = TagStates.WAIT_END

        self.stop()

    def update_state(self, state):
        # after MOH was done, continue to the next state
        if state == TagStates.WAIT_MOH:
            state = self.last_state

        # when event occurred
        if self.stop_app_event.is_set():
            if state == TagStates.TAG_TEST or state == TagStates.PROCESS_TEST:
                pass  # finish the cycle
            else:
                self.last_state = state
                state = TagStates.STOP
        elif self.moh.get_manual_operation_is_needed():
            self.last_state = state
            state = TagStates.WAIT_MOH

        # main flow
        if state == TagStates.IDLE:
            state = TagStates.GET_TRIGGER
        elif state == TagStates.GET_TRIGGER:
            state = TagStates.START_TEST
        elif state == TagStates.START_TEST:
            state = TagStates.TAG_TEST
        elif state == TagStates.TAG_TEST:
            state = TagStates.PROCESS_TEST
        elif state == TagStates.PROCESS_TEST:
            state = TagStates.END_TEST
        elif state == TagStates.END_TEST:
            state = TagStates.WAIT_END
        elif state == TagStates.WAIT_END:
            state = TagStates.NEXT_TAG
        elif state == TagStates.NEXT_TAG:
            state = TagStates.GET_TRIGGER

        return state

    def wait_for_new_event(self, event, time_to_wait=TagTestingDefaults.POST_RUN_WAIT_TIME):
        t_start = time.time()
        while not event.is_set():
            if self.stop_app_event.is_set():
                self.logger.info('Tag Testing: stop application during waiting for event')
                time.sleep(TagTestingDefaults.POST_RUN_WAIT_TIME)
                break
            elif self.moh.get_manual_operation_is_needed():
                self.wait_for_moh()
                t_start = time.time()
            event.wait(0.1)
            dt = time.time() - t_start

            if dt > time_to_wait and not self.moh.get_manual_operation_is_needed():
                raise TagTestingException('wait_for_new_event: did not receive ack for '
                                          'starting/ending tag test')
        event.clear()

    def update_missing_label_parameters(self):
        self.missing_labels += 1
        if self.user_config['sensorOffset'] != '':
            self.missing_label_locations.append(self.tested + int(self.user_config['sensorOffset']) + 1)
            self.logger.info(f'TagTesting: update_missing_label_parameters: '
                             f'estimate Missing Label at location: {self.missing_label_locations[-1]}')

    def get_trigger(self):
        pulse_received = self.wiliot_tester.wait_for_trigger(
            wait_for_gw_trigger=TagTestingDefaults.TIMEOUT_FOR_MISSING_LABEL)
        if not pulse_received:
            while self.moh.get_manual_operation_is_needed():
                self.wait_for_moh()
                if self.stop_app_event.is_set():
                    return
                pulse_received = self.wiliot_tester.wait_for_trigger(
                    wait_for_gw_trigger=TagTestingDefaults.TIMEOUT_FOR_MISSING_LABEL)
                if pulse_received:
                    break
            if not pulse_received:
                self.update_missing_label_parameters()
                if TagTestingDefaults.ENABLE_MISSING_LABEL and \
                        self.missing_labels < TagTestingDefaults.MAX_MISSING_LABEL_ENGINEERING + 1:
                    self.need_to_manual_trigger = True
                    self.logger.info((f'MISSING LABEL. no trigger was received for '
                                      f'{TagTestingDefaults.TIMEOUT_FOR_MISSING_LABEL} seconds. Continue to test.'))
                    return
                if not self.stop_app_event.is_set():
                    raise MissingLabelException(f'MISSING LABEL. no trigger was received for '
                                                f'{TagTestingDefaults.TIMEOUT_FOR_MISSING_LABEL} seconds')

    def start_test(self):
        self.test_start_event.set()
        if self.attenuator_handle.enable:
            self.attenuator_handle.set_dynamic_value()
        if self.is_r2r:
            if self.test.get_total_test_time() > TagTestingDefaults.MAX_R2R_WAIT_TIME:
                self.r2r.send_stop_to_r2r()

    def gw_reset_and_config(self):
        self.wiliot_tester.gw_reset_and_config()
        self.wiliot_tester.init_gw_test(is_start_gw_app=False)

    def tag_test(self):
        self.tested += 1
        self.tag_test_running = True
        if not self.is_printing_calibration():
            tester_res = self.wiliot_tester.run(wait_for_gw_trigger=None,
                                                need_to_manual_trigger=self.need_to_manual_trigger)
        else:
            tester_res = WiliotTesterTagResultList()
            time.sleep(TagTestingDefaults.TIME_BETWEEN_TEST_PRINTING)
        self.tag_test_running = False

        return tester_res

    def is_test_running(self):
        return self.tag_test_running

    def is_correct_tag_gen(self, adva_in):
        self.is_gen_checked = True
        if len(adva_in) < 2:
            return True
        if adva_in.lower()[:2] >= '06':
            return self.user_config['gen'].lower() == 'gen3'
        return self.user_config['gen'].lower() == 'gen2'   # adva_in.lower()[:2] < '06'

    def process_test(self, tester_res):
        is_pass = False
        tag_current_location = self.tested - 1
        if self.is_printing_calibration():
            tester_res.set_total_test_status(True)
            is_pass = True
        elif tester_res.is_all_tests_passed():
            selected_tag = tester_res.check_and_get_selected_tag_id()
            if selected_tag == '':
                if not self.wiliot_tester.run_all and not self.is_printing_calibration():
                    raise TagTestingException("run: ANALYSIS TEST: Test Status is PASS but could not select tag")
                else:
                    is_pass = True

            elif selected_tag in self.all_selected_tags:
                # Duplication
                tester_res.set_total_fail_bin(FailureCodes.DUPLICATION_OFFLINE)
                tester_res.set_packet_status(adv_address=selected_tag, status='duplication')
                self.all_duplicated_tags.append(selected_tag)
                self.logger.warning(f'DUPLICATION for Adva: {selected_tag}')
            else:
                # PASS
                self.all_selected_tags.append(selected_tag)
                is_pass = True

            if not self.is_gen_checked and not self.is_correct_tag_gen(selected_tag):
                raise Exception(f'Mismatch between tag gen: {selected_tag} and '
                                f'specified gen: {self.user_config["gen"]}')
        else:
            if tester_res.get_total_fail_bin() == FailureCodes.NO_RESPONSE.value:
                self.logger.info(f'TagTesting: process_test: got NO_RESPONSE at location: {tag_current_location}')
                if self.tested in self.missing_label_locations:
                    tester_res.set_total_fail_bin(FailureCodes.MISSING_LABEL, overwrite=True)
                    self.logger.info(f'set tag location {tag_current_location} to MISSING lABEL fail bin')
        if not is_pass:
            self.logger.warning(f'Tag {tag_current_location} Failed - {tester_res.get_total_fail_bin(as_name=True)}')
        self.all_tags += tester_res.get_test_unique_adva()
        self.all_tags = list(set(self.all_tags))
        self.is_passed_list.append(is_pass)

    def get_all_duplicated_tags(self):
        return self.all_duplicated_tags

    def is_printing_calibration(self):
        return self.user_config['printingFormat'].lower() == 'test' and self.user_config['toPrint'].lower() == 'yes'

    @staticmethod
    def calculating_ttfp_avg(ttfp_list):
        ttfp_list_no_nan = [x for x in ttfp_list if not np.isnan(x)]
        if ttfp_list_no_nan:
            return np.mean(ttfp_list_no_nan)
        return float(-1)

    def end_test(self, tester_res):
        self.ttfp_list.append(tester_res.get_total_ttfp())
        test_data = {'tested': self.get_tested(),
                     'passed': self.get_passed(),
                     'responded': self.get_responded(),
                     'missing_label': self.get_missing_label_count(),
                     'ttfp_avg': self.calculating_ttfp_avg(self.ttfp_list),
                     'crc_environment_previous': self.wiliot_tester.get_wrong_crc()
                     }
        if self.sensors_handle.enable:
            try:
                sensors_out = self.sensors_handle.get_measurement()
                test_data = {**test_data, **sensors_out}
            except Exception as e:
                self.logger.warning(f'end test: could not read from sensor due to {e}')
        tester_res.set_test_info(test_info=test_data)
        if self.tag_results_q.full():
            raise TagTestingException('end_test: tag results queue is full - error in offset definition')
        self.tag_results_q.put(tester_res, block=False)
        self.test_end_event.set()

    def wait_for_moh(self):
        self.logger.info('TagTesting: wait for manual operation handling')
        try:
            if self.is_r2r:
                self.r2r.send_stop_to_r2r()
        except Exception as e:
            self.logger.warning(f'Could not stop R2R during wait for manual operation handling due to {e}')
        while True:
            time.sleep(1)
            if not self.moh.get_manual_operation_is_needed():
                break
            if self.stop_app_event.is_set():
                break

    def get_tested(self):
        return self.tested

    def get_responded(self):
        return len(self.all_tags)

    def get_passed(self):
        return len(self.all_selected_tags)

    def get_missing_label_count(self):
        return self.missing_labels

    def get_is_passed_list(self):
        return self.is_passed_list.copy()

    def get_ttfp_list(self):
        return self.ttfp_list

    def stop(self):
        if self.is_r2r:
            self.post_run_validation()
            self.r2r.exit()
        self.wiliot_tester.exit_tag_test()
        self.logger.info('TagTesting Thread is done')

    def post_run_validation(self):
        self.logger.info('TagTesting: start post run validation...')
        if self.last_state == TagStates.WAIT_END:
            stage = 'move'  # wait, move, trigger
        elif self.last_state == TagStates.GET_TRIGGER or self.last_state == TagStates.STOP:
            stage = 'trigger'  # wait, move, trigger
        else:
            stage = 'wait'  # wait, move, trigger
        while True:
            if stage == 'wait':
                time.sleep(TagTestingDefaults.TIME_BETWEEN_TEST_PRINTING)
                self.validation_end_event.wait(TagTestingDefaults.POST_RUN_WAIT_TIME)
                if not self.validation_end_event.is_set():
                    break
                stage = 'move'
            elif stage == 'move':
                self.validation_end_event.clear()
                self.r2r.move_r2r()
                gw_answer = self.gw_obj.read_specific_message(msg="Start Production Line GW",
                                                              read_timeout=TagTestingDefaults.EVENT_WAIT_TIME)
                self.logger.info(f'TagTesting: post_run_validation: gw rsp: {gw_answer}')
                stage = 'trigger'

            elif stage == 'trigger':
                self.test_start_event.set()
                self.gw_obj.stop_gw_app()
                self.test_end_event.set()
                stage = 'wait'

            else:
                raise Exception(f'TagTesting: post_run_validation: stage is not supported: {stage}')

        self.logger.info('TagTesting: end post run validation')

    def get_gw_version(self):
        return self.wiliot_tester.get_gw_version()

    def get_gw_gpio_signal(self):
        signals_out = []
        t_i = datetime.datetime.now()
        dt = 0.0
        while dt < TagTestingDefaults.PRINTER_SIGNAL_TIMEOUT:
            try:
                time.sleep(TagTestingDefaults.PRINTER_SIGNAL_SLEEP_TIME)
                all_signals = self.gw_obj.get_gw_signals()
                if all_signals:
                    signals_out = [sig['raw'] for sig in all_signals]
                    self.logger.info(f'got the following signals: {signals_out}')
                    break
            except Exception as e:
                self.logger.warning(f'get_gw_gpio_signal got exception, try again ({e})')
                time.sleep(TagTestingDefaults.PRINTER_SIGNAL_SLEEP_TIME)
            dt = (datetime.datetime.now() - t_i).total_seconds()
        return signals_out


class AttenuatorHandle(object):
    def __init__(self, logger_name, exception_q, attenuator_configs=None):
        self.logger = logging.getLogger(logger_name)
        self.exception_q = exception_q
        self.enable = True
        if attenuator_configs is None or attenuator_configs['AutoAttenuatorEnable'].lower() == 'no':
            self.logger.info('Attenuator is disable')
            self.enable = False
            return

        self.attenuator_configs = attenuator_configs
        if self.attenuator_configs['attnComport'].upper() == 'AUTO':
            self.attenuator = Attenuator('API').GetActiveTE()
        else:
            self.attenuator = Attenuator('API', comport=f'COM{self.attenuator_configs["attnComport"]}').GetActiveTE()
        current_attn = self.attenuator.Getattn()
        self.logger.info(f'Attenuator is connected at port {self.attenuator.comport} and set to: {current_attn}')
        self.set_value()

    def set_value(self, attenuator_val=None):
        if attenuator_val is None:
            attenuator_val = self.attenuator_configs['attnval']
        try:
            set_attn_status = self.attenuator.Setattn(int(attenuator_val))
        except Exception as e:
            raise AttenuatorException(f'AttenuatorHandle Exception: set_value: {e}')
        if set_attn_status == attenuator_val:
            self.logger.info(f'Attenuation is set to {attenuator_val} dB')
        else:
            raise AttenuatorException(f'AttenuatorHandle Exception: set_value: failed to set attenuation value '
                                      f'(expected: {attenuator_val}, current: {set_attn_status})')

    def set_dynamic_value(self):
        pass


class ExternalSensorHandle(object):
    def __init__(self, logger_name, exception_q, sensor_configs=None):
        self.logger = logging.getLogger(logger_name)
        self.exception_q = exception_q
        self.sensor_configs = sensor_configs

        self.enable = True
        if sensor_configs is None or self.sensor_configs['sensorsEnable'].lower() == 'no':
            self.logger.info('External Sensors measurement is disable')
            self.enable = False
            return

        self.sensor = None
        self.sensor_measurements = []
        try:
            self.connect()
        except Exception as e:
            raise SensorException(f'ExternalSensorHandle Exception: init: {e}')

    def connect(self):
        self.sensor = YoctoSensor(self.logger)
        if self.sensor.temperature_sensor is not None:
            self.logger.info(f'Connected to temperature sensor')
            self.sensor_measurements.append({'name': 'temperature', 'func': self.sensor.get_temperature})
        if self.sensor.humidity_sensor is not None:
            self.logger.info(f'Connected to humidity sensor')
            self.sensor_measurements.append({'name': 'humidity', 'func': self.sensor.get_humidity})
        if self.sensor.light_sensor is not None:
            self.logger.info(f'Connected to light sensor')
            self.sensor_measurements.append({'name': 'light', 'func': self.sensor.get_light})

    def get_measurement(self):
        meas_out = {}
        for meas in self.sensor_measurements:
            try:
                cur_value = meas['func']()
                if cur_value == float('nan') or (isinstance(cur_value, int) and cur_value == 0):
                    self.logger.warning(f'ExternalSensorHandle: Could not measure {meas["name"]} from sensor')
                else:
                    self.logger.info(f'Measured {meas["name"]} from sensor: {cur_value}')
                    meas_out[f'{meas["name"]}_sensor'] = cur_value
            except Exception as e:
                raise SensorException(f'ExternalSensorHandle: get_measurement: {e}')
        return meas_out



import pandas as pd
import time


def get_last_pass_external_id(packet_data_path, df_packet=None):
    df_packet = pd.read_csv(packet_data_path) if df_packet is None else df_packet
    last_location = max(df_packet[df_packet['status_offline'] == 1]['tag_run_location'])
    external_id = df_packet[df_packet['tag_run_location'] == last_location]['external_id'].unique()
    if len(external_id) != 1:
        raise Exception(f'Could not extract last external id since external id is not specified or several external ids'
                        f' on the same location: {external_id}')
    return external_id[0]


def delete_entry_from_specific_external_id(packet_data_path, run_data_path, external_id, new_file_suffix='_edit'):
    df_packet = pd.read_csv(packet_data_path)
    df_run = pd.read_csv(run_data_path)

    i_to_remove = df_packet[df_packet['external_id'] == external_id].index
    if len(i_to_remove) == 0:
        raise Exception('could not find the specifies external id')
    df_packet_edit = df_packet.drop(list(range(min(i_to_remove), len(df_packet))))

    if df_packet_edit.empty:
        df_run['total_run_tested'] = 0
        df_run['total_run_passed_offline'] = 0
        df_run['total_run_responding_tags'] = 0
    else:
        df_run['total_run_tested'] = \
            len(df_packet_edit['tag_run_location'][df_packet_edit['fail_bin_str'] != 'BAD_PRINTING'].unique())
        df_run['total_run_passed_offline'] = \
            len(df_packet_edit[df_packet_edit['status_offline'] == 1]['tag_run_location'].unique())
        df_run['total_run_responding_tags'] = \
            len(df_packet_edit['adv_address'][df_packet_edit['packet_status'] == 'good'].unique())
    cur_comments = df_run['comments'][0] if not pd.isnull(df_run['comments'][0]) else ''
    last_external_id = get_last_pass_external_id(packet_data_path, df_packet)
    df_run['comments'] = cur_comments + f'.del:{external_id[-4:]}-{last_external_id[-4:]}' \
                                        f'@{time.strftime("%d%m%y_%H%M%S")}'
    packet_data_path = packet_data_path.replace('.CSV', '.csv')
    run_data_path = run_data_path.replace('.CSV', '.csv')
    df_packet_edit.to_csv(packet_data_path.replace('.csv', f'{new_file_suffix}.csv'), index=False)
    df_run.to_csv(run_data_path.replace('.csv', f'{new_file_suffix}.csv'), index=False)
    return df_run, df_packet_edit


def update_test_vars(self, new_df, df_packet_edit):
    self.test_data['tag_run_location'] = new_df['total_run_tested'][0]
    self.run_data['total_run_responding_tags'] = new_df['total_run_responding_tags'][0]
    self.run_data['total_run_tested'] = new_df['total_run_tested'][0]
    self.run_data['total_run_passed_offline'] = new_df['total_run_passed_offline'][0]
    self.run_data['comments'] = new_df['comments'][0]
    self.all_tags = df_packet_edit['adv_address'].unique().tolist()
    self.all_selected_tags = df_packet_edit['selected_tag'].unique().tolist()


if __name__ == '__main__':
    FILE_PATH = 'C:/Users/shunit/eclipse-workspace/post_process_testing/offline/duplication1'
    commmon_run_name = '02jx_20221224_203102'
    packet_data_path = f"{FILE_PATH}/{commmon_run_name}@packets_data.csv"
    run_data_path = f"{FILE_PATH}/{commmon_run_name}@run_data.csv"

    last_external_id = get_last_pass_external_id(packet_data_path)
    print(last_external_id)

    external_id_to_delete = last_external_id
    new_df = delete_entry_from_specific_external_id(packet_data_path, run_data_path,
                                                    external_id=external_id_to_delete, new_file_suffix='_edit')
    print(new_df)
    print('done')



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
import json
import logging
import threading
import multiprocessing
import time
import datetime
import pandas as pd

from wiliot_tools.association_tool.association_configs import is_asset_code, is_wiliot_code, \
    WILIOT_MIN_NUM_CODE, ASSET_NUM_CODES
from wiliot_tools.association_tool.send_association_to_cloud import CloudAssociation
from wiliot_testers.association_tester.hw_components.r2r_component import R2R
from wiliot_testers.association_tester.hw_components.scanner_component import Scanner
pd.options.mode.chained_assignment = None  # default='warn'


ASSOCIATION_Q_SIZE = 200


class AssociatorProcess(CloudAssociation):
    def __init__(self, associate_q, associated_status_q, stop_event, owner_id, category_id, time_btwn_request=1):
        self.associated_status_q = associated_status_q
        super().__init__(associate_q=associate_q, stop_event=stop_event, owner_id=owner_id,
                         is_gcp=False, category_id=category_id, time_btwn_request=time_btwn_request)

    def handle_results(self, message, asset_id, pixel_dict, bad_association):
        bad_association = super().handle_results(message, asset_id, pixel_dict, bad_association)
        self.associated_status_q.put({'is_associated': int(message['status_code'] // 100) == 2,
                                      'associate_status_code': message['status_code'],
                                      'asset_code': [asset_id],
                                      'wiliot_code': pixel_dict['pixel_id']})
        return bad_association


class TagAssociation(object):
    def __init__(self, user_input, stop_event, is_app_running, logger_name, logger_path, exception_q, run_param_file):
        self.logger = logging.getLogger(logger_name)
        self.logger_path = logger_path
        self.user_inputs = user_input
        self.stop_event = stop_event
        self.exception_q = exception_q
        self.is_running = False
        self.is_main_app_running = is_app_running
        self.run_param_file = run_param_file

        self.r2r = None
        self.scanner = None
        self.init_hw()

        # init data:
        self.locations_df = pd.DataFrame()
        # start app
        self.associator_handler = None
        try:
            need_to_associate_q = multiprocessing.Queue(maxsize=ASSOCIATION_Q_SIZE)
            associated_q = multiprocessing.Queue(maxsize=ASSOCIATION_Q_SIZE)

            if self.user_inputs['need_to_associate'].lower() == 'yes':
                self.associator_handler = multiprocessing.Process(target=AssociatorProcess,
                                                                  args=(need_to_associate_q,
                                                                        associated_q,
                                                                        stop_event,
                                                                        self.user_inputs['owner_id'],
                                                                        self.user_inputs['category_id']
                                                                        ))
            self.need_to_associate_q = need_to_associate_q
            self.associated_q = associated_q

        except Exception as e:
            self.logger.warning(f'exception during TagAssociation init: {e}')
            raise e

    def init_hw(self):
        # connect to the r2r/arduino:
        self.r2r = R2R(logger_name=self.logger.name, counter_start_idx=self.user_inputs['first_location'])
        # connect to scanner
        self.scanner = Scanner(logger_name=self.logger.name, max_test_time=self.user_inputs['max_test_time'])

    def run_app(self):
        try:
            if self.associator_handler is not None:
                self.associator_handler.start()
        except Exception as e:
            self.logger.warning(f'exception during TagAssociation run_app: {e}')
            raise e
        self.run()

    def run(self):
        self.logger.info('MoveAndScan Start')
        self.is_running = True
        while True:
            time.sleep(0)
            cur_time = time.time()
            try:
                if self.stop_event.is_set():
                    self.logger.info('MoveAndScan Stop')
                    self.exit_app()
                    return
                elif self.is_running != self.is_main_app_running():
                    self.is_running = not self.is_running
                    if self.is_running:
                        self.continue_app()
                    else:
                        self.pause_app()
                elif not self.exception_q.empty():
                    time.sleep(0.1)
                    continue

                if self.is_running:
                    # scan
                    res = self.scan()

                    # add data to dataframe:
                    self.add_data(new_result=res)

                    # wait
                    time_to_wait = max([0, float(self.user_inputs['min_test_time']) - (time.time() - cur_time)])
                    time.sleep(time_to_wait)

                    # move
                    if self.user_inputs['is_step_machine'].lower() == 'yes':
                        self.r2r.move()
                        time.sleep(float(self.user_inputs['time_to_move']))

                    if not self.associated_q.empty():
                        self.merge_associated_data()

                else:
                    time.sleep(1)

            except Exception as e:
                self.logger.warning(f'MoveAndScan got exception: {e}')
                if not self.exception_q.full():
                    self.exception_q.put(f'MoveAndScan: {e}')

    def add_data(self, new_result):
        new_result['wiliot_code'] = new_result['wiliot_code'] if len(new_result['wiliot_code']) > 0 else ['']
        new_result['asset_code'] = new_result['asset_code'] if len(new_result['asset_code']) > 0 else ['']
        added_data = []
        for wiliot_code in new_result['wiliot_code']:
            for asset_code in new_result['asset_code']:
                new_row = {k: v for k, v in new_result.items()}
                new_row['wiliot_code'] = wiliot_code
                new_row['asset_code'] = asset_code
                added_data.append(new_row)

        self.update_last_run_params(added_data[-1])
        self.locations_df = pd.concat([self.locations_df, pd.DataFrame(added_data)])

    def update_last_run_params(self, new_data):
        with open(self.run_param_file, 'rb') as f:
            last_run_params = json.load(f)
        last_run_params['last_location'] = new_data['location']
        last_run_params['last_run_name'] = self.user_inputs['run_name']
        if new_data['asset_code']:
            last_run_params['last_asset_id'] = new_data['asset_code']
            last_run_params['last_asset_location'] = new_data['location']
        with open(self.run_param_file, 'w') as f:
            json.dump(last_run_params, f)

    def scan(self):
        complete_label_read = {'location': self.r2r.get_counter(),
                               'wiliot_code': [], 'asset_code': [], 'timestamp': 0,
                               'scan_status': False,
                               'is_associated': False, 'associate_status_code': ''}

        scanned_codes = self.scanner.scan()

        self.check_scanned_codes_and_update_result(scanned_codes, complete_label_read)

        # send data to cloud
        if self.user_inputs['need_to_associate'].lower() == 'yes' and complete_label_read['scan_status']:
            if self.need_to_associate_q.full():
                self.logger.warning(f'need_to_associate_q queue is full. discard  {complete_label_read}')
            else:
                self.need_to_associate_q.put(complete_label_read)

        # end of label scanning:
        self.check_end_of_label(complete_label_read)
        return complete_label_read

    def check_scanned_codes_and_update_result(self, codes_in, complete_label_read):
        for code in codes_in:
            if is_wiliot_code(code):
                complete_label_read['wiliot_code'].append(code)
                complete_label_read['timestamp'] = datetime.datetime.now().timestamp()
            elif is_asset_code(code):
                complete_label_read['asset_code'].append(code)
                complete_label_read['timestamp'] = datetime.datetime.now().timestamp()

        if len(complete_label_read['wiliot_code']) >= WILIOT_MIN_NUM_CODE \
                and len(complete_label_read['asset_code']) == ASSET_NUM_CODES:
            if (self.user_inputs['asset_location'] == 'first' and is_wiliot_code(codes_in[-1])) or (
                    self.user_inputs['asset_location'] == 'last' and is_wiliot_code(codes_in[0])):
                complete_label_read['scan_status'] = True
                self.logger.info(f'Found all codes for association! '
                                 f'Wiliot:{complete_label_read["wiliot_code"]}, '
                                 f'Asset: {complete_label_read["asset_code"]}')
            else:
                complete_label_read['scan_status'] = False
                raise Exception(f'Scanned codes from different labels '
                                f'Wiliot:{complete_label_read["wiliot_code"]}, '
                                f'Asset: {complete_label_read["asset_code"]}')

    def check_end_of_label(self, label_read):
        if label_read is None:
            return
        if len(label_read['asset_code']) == 0 and len(label_read['wiliot_code']) == 0:
            self.logger.info(f"No codes were read")
        if len(label_read['asset_code']) < ASSET_NUM_CODES:
            self.logger.info(f"Not enough Asset codes were scanned: {label_read['asset_code']}")
        elif len(label_read['asset_code']) > ASSET_NUM_CODES:
            self.logger.info(f"Too many Asset codes were scanned: {label_read['asset_code']}")
        elif len(label_read['wiliot_code']) < WILIOT_MIN_NUM_CODE:
            self.logger.info(f"Not enough Wiliot codes were scanned: {label_read['wiliot_code']}")

    def merge_associated_data(self):
        n = self.associated_q.qsize()
        for _ in range(n):
            associated = self.associated_q.get(block=False)
            print(f'associated: {associated}')
            self.locations_df.loc[(self.locations_df['asset_code'].isin(associated['asset_code'])) & (
                    self.locations_df['wiliot_code'].isin(associated['wiliot_code'])),
                                  ['is_associated', 'associate_status_code']] = [associated['is_associated'],
                                                                                 associated['associate_status_code']]

    def pause_app(self):
        self.r2r.disconnect()
        self.scanner.disconnect()

    def continue_app(self):
        self.r2r.connect()
        self.scanner.reconnect()

    def exit_app(self):
        try:
            self.r2r.disconnect()
            self.scanner.disconnect()
            if self.associator_handler is not None:
                self.associator_handler.join(10)
                if self.associator_handler.is_alive():
                    self.logger.warning('associator process is still running')
            if not self.associated_q.empty():
                self.merge_associated_data()
        except Exception as e:
            self.logger.warning(f'MoveAndScan: exit_app: got exception: {e}')
            if not self.exception_q.full():
                self.exception_q.put(f'MoveAndScan: exit_app: {e}')

    def get_locations_df(self):
        return self.locations_df


if __name__ == '__main__':
    from wiliot_core import set_logger
    RUN_TIME = 60

    tag_assoc_logger_path, tag_assoc_logger = set_logger(app_name='TagAssociation', dir_name='tag_association',
                                                         file_name='association_log')
    stop_event = multiprocessing.Event()
    user_input = {
        'min_test_time': '1.5',
        'max_test_time': '5',
        'time_to_move': '0.5',
        'need_to_associate': 'no',
        'is_step_machine': 'yes',
        'asset_location': 'last',
        'owner_id': 'wiliot-ops',
        'category_id': '86fd9f07-38b0-466f-b717-2e285b803f5c'
    }
    ta = TagAssociation(user_input=user_input, stop_event=stop_event,
                        logger_name=tag_assoc_logger.name, logger_path=tag_assoc_logger_path)

    t_i = time.time()
    while time.time() - t_i < RUN_TIME:
        time.sleep(1)
        df = ta.get_locations_df()
        print(f'n unique locations: {len(df)}')
    # stop run
    stop_event.set()

    df = ta.get_locations_df()
    df_path = tag_assoc_logger_path.replace('.log', '_locations_df.csv')
    print(f'saving data at: {df_path}')
    df.to_csv(df_path, index=False)

    print('done')



"""
  Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

from queue import Queue
import threading
import pandas as pd

from wiliot_testers.sample.modules.wiliot_sample_tag_test import WiliotSampleTagTest, TesterName, TagStatus
from wiliot_tools.resolver_tool.resolve_packets import ResolvePackets


class SampleRun(object):
    def __init__(self, tags_in_test, selected_test, all_configs, cloud_config,
                 gw_obj=None, stop_event=None, inlay=None, logger_name=None):
        resolve_q = Queue(maxsize=100) if cloud_config['is_online'] else None
        self.stop_event = stop_event if stop_event is not None else threading.Event()
        self.stop_event.clear()
        self.results_df = None

        self.sample_tester = WiliotSampleTagTest(tags_in_test=tags_in_test,
                                                 selected_test=selected_test,
                                                 test_suite=all_configs,
                                                 gw_obj=gw_obj,
                                                 stop_event_trig=self.stop_event,
                                                 tester_name=TesterName.SAMPLE,
                                                 logger_name=logger_name,
                                                 inlay=inlay,
                                                 resolve_q=resolve_q)
        self.logger = self.sample_tester.logger
        self.resolver = ResolvePackets(tags_in_test=tags_in_test,
                                       owner_id=cloud_config['owner_id'],
                                       env=cloud_config['env'],
                                       resolve_q=resolve_q,
                                       set_tags_status_df=self.sample_tester.set_tags_status,
                                       stop_event_trig=self.stop_event,
                                       logger_name=self.logger.name,
                                       gui_type='ttk',
                                       tag_status=TagStatus
                                       ) if cloud_config['is_online'] else None

    def run(self):
        self.logger.info('start running sample test thread')
        sample_tester_handler = threading.Thread(target=self.sample_tester.run, args=())
        sample_tester_handler.start()
        if self.resolver is not None:
            self.logger.info('start running resolver thread')
            resolver_handler = threading.Thread(target=self.resolver.run, args=())
            resolver_handler.start()
        else:
            resolver_handler = None

        sample_tester_handler.join()
        self.logger.info('running sample test thread was done')
        if resolver_handler is not None:
            self.stop_event.set()
            resolver_handler.join(10)
            if resolver_handler.is_alive():
                raise Exception('could not stop the resolve thread, please re-run app')
            self.logger.info('running resolver thread was done')

        self.logger.info('end of run')

    def get_test_results(self, only_inside_test=False):
        if self.results_df is not None:
            return self.results_df
        test_results = self.sample_tester.test_results
        df = test_results.tests[0].filtered_tags.get_df(add_sprinkler_info=True)
        if df.empty:
            self.results_df = df
            return df
        stat = test_results.tests[0].filtered_tags.get_statistics()
        df_res = pd.merge(df, self.sample_tester.tags_status_df, how='outer', on='adv_address')
        df_res = pd.merge(df_res, stat, how='outer', on='adv_address')
        if only_inside_test or self.sample_tester.is_all_resolve():
            df_res = df_res[df_res['resolve_status'] == TagStatus.INSIDE_TEST]
        df_res.reset_index(inplace=True)
        self.results_df = df_res
        return df_res


if __name__ == '__main__':
    import os
    import json

    test_tags_in_test = ['(01)00850027865010(21)01t6T0080', '(01)00850027865010(21)01jvT5192']
    test_selected_test = 'TIKI_BLE'
    test_configs_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'configs', '.default_test_configs.json')
    with open(test_configs_path, 'r') as f:
        test_configs = json.load(f)
    test_all_configs = {'run': {'rssiThresholdHW': 0,
                                'rssiThresholdSW': 100,
                                },
                        'test': test_configs
                        }
    test_cloud_config = {'is_online': True, 'owner_id': 'wiliot-ops', 'env': 'prod'}

    s_r = SampleRun(tags_in_test=test_tags_in_test,
                    selected_test=test_selected_test,
                    all_configs=test_all_configs,
                    cloud_config=test_cloud_config,
                    gw_obj=None,
                    stop_event=None,
                    inlay=None,
                    logger_name=None)
    s_r.run()
    res = s_r.get_test_results()
    print(res)
    print('done')



"""
  Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import os
import datetime
import logging

from wiliot_core import InlayTypes, WiliotDir

from wiliot_testers.wiliot_tester_log import dict_to_csv, WiliotTesterLog
from wiliot_testers.wiliot_tester_tag_test import TesterName
from wiliot_testers.wiliot_tester_tag_result import ConversionTypes, SurfaceTypes, FailureCodes
from wiliot_testers.utils.get_version import get_version
from wiliot_testers.offline.modules.offline_printing_and_validation import ValidatedBin


class DataLogging(object):
    def __init__(self, exception_queue, log_config=None):
        self.exception_queue = exception_queue

        # init log config
        self.log_config = log_config
        self.log_obj = None
        self.run_data_path = ''
        self.packets_data_path = ''
        self.reel_name = ''
        self.data_folder = ''
        logger_name = self.set_log_data_dir()
        self.logger = logging.getLogger(logger_name)

        self.run_data = {}
        self.test_data = {}
        self.packet_headers_default = []
        self.tag_reel_location = 0
        self.total_tested = 0

    def get_data_folder(self):
        return self.data_folder

    def get_test_folder(self):
        return self.log_obj.data_folder

    def get_logger_name(self):
        return self.logger.name

    def get_results_logger_name(self):
        return self.log_obj.results_logger.name

    def get_gw_logger_name(self):
        return self.log_obj.gw_logger.name

    def set_log_data_dir(self):
        wiliot_dir = WiliotDir()
        app_dir = wiliot_dir.get_wiliot_root_app_dir()

        log_dir = os.path.join(app_dir, 'offline', 'logs', self.log_config['batchName'])
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)

        self.data_folder = log_dir
        self.reel_name = self.log_config['batchName'].rstrip()
        run_start_time = datetime.datetime.now()
        common_run_name = self.reel_name + '_' + run_start_time.strftime("%Y%m%d_%H%M%S")

        self.log_obj = WiliotTesterLog()
        self.log_obj.set_common_run_name(reel_name=self.reel_name, run_start_time=run_start_time)
        self.log_obj.set_logger(log_path=log_dir, tester_name='offline')
        self.run_data_path, self.packets_data_path = self.log_obj.create_data_dir(data_path=log_dir,
                                                                                  tester_name='offline',
                                                                                  run_name=common_run_name)
        if 'testerStationName' not in os.environ:
            raise Exception('testerStationName is missing from PC environment variables, '
                            'please add it in the following convention: <company name>_<tester number>')
        self.log_obj.set_station_name(os.environ['testerStationName'])

        return self.log_obj.logger.name

    def update_log_config(self, log_config):
        for k, v in log_config.items():
            self.log_config[k] = v

    def get_log_config(self):
        return self.log_config

    def get_run_data_path(self):
        return self.run_data_path

    def get_packets_data_path(self):
        return self.packets_data_path

    def get_reel_name(self):
        return self.reel_name

    def run_data_init(self, is_printing, gw_version, test_suite_dict):
        common_run_name = self.log_obj.run_name
        run_start_time = self.log_obj.run_start_time

        if 'tag_reel_location' in self.log_config.keys():
            self.tag_reel_location = self.log_config['tag_reel_location']

        self.run_data = {'common_run_name': common_run_name,
                         'tester_station_name': self.log_obj.tester_station_name,
                         'operator': self.log_config['operator'], 'reel_run_start_time': run_start_time,
                         'reel_run_end_time': None,
                         'batch_name': self.log_config['batchName'],
                         'tester_type': TesterName.OFFLINE.value,
                         'comments': self.log_config['comments'],
                         'total_run_tested': 0, 'total_run_responding_tags': 0,
                         'total_run_passed_offline': 0,
                         'total_run_bad_printing': 0,
                         'total_missing_labels': 0, 'run_responsive_tags_yield': float('nan'),
                         'run_offline_yield': float('nan'),
                         'conversion_type': ConversionTypes(self.log_config['conversion']).name,
                         'inlay': InlayTypes(self.log_config['inlay']).name,
                         'test_suite': self.log_config['testName'],
                         'test_suite_dict': test_suite_dict,
                         'surface': SurfaceTypes(self.log_config['surface']).name,
                         'to_print': self.log_config['toPrint'],
                         'qr_validation': self.log_config['QRRead'],
                         'qr_offset': self.log_config['QRoffset'],
                         'printing_offset': self.log_config['printOffset'],
                         'print_pass_job_name': self.log_config['passJobName'] if is_printing else '',
                         'printing_format': self.log_config['printingFormat'] if is_printing else '',
                         'external_id_prefix': self.log_config['stringBeforeCounter'] if is_printing else '',
                         'external_id_suffix_init_value': self.log_config['firstPrintingValue'] if is_printing else '',
                         'gw_version': gw_version,
                         'py_wiliot_version': get_version(),
                         'owner_id': 'wiliot-ops',
                         'sensors_enable': self.log_config['sensorsEnable'],
                         'ttfp_avg': float('nan'),
                         'user_config_dict': self.log_config,
                         'conversion_label': self.log_config['conversionLabel']}

        self.test_data = {'common_run_name': common_run_name, 'tag_run_location': 0,
                          'total_test_duration': None,
                          'status_offline': False, 'fail_bin': FailureCodes.NONE.value,
                          'fail_bin_str': FailureCodes.NONE.name,
                          'external_id': '',
                          'label_validated': ValidatedBin.UNTESTED.name, 'trigger_time': None,
                          'test_num': 0,
                          'tag_reel_location': self.tag_reel_location}

        self.packet_headers_default = \
            ['common_run_name', 'tag_run_location', 'tag_reel_location', 'test_num', 'external_id', 'time_from_start',
             'raw_packet', 'rssi',
             'packet_status', 'adv_address', 'selected_tag', 'is_test_pass', 'status_offline', 'fail_bin',
             'fail_bin_str',
             'test_status', 'num_packets', 'num_cycles', 'sprinkler_counter_mean', 'sprinkler_counter_std',
             'sprinkler_counter_min', 'sprinkler_counter_max', 'tbp_mean', 'tbp_std', 'tbp_min', 'tbp_max',
             'tbp_num_vals',
             'per_mean', 'per_std', 'rssi_mean', 'rssi_std', 'rssi_min', 'rssi_max', 'ttfp', 'rx_rate_normalized',
             'rx_rate', 'total_test_duration', 'test_start_time', 'trigger_time',
             'test_end_time', 'label_validated', 'rx_channel', 'energizing_pattern', 'time_profile',
             'decrypted_packet_type',
             'group_id', 'flow_ver', 'test_mode', 'en', 'type', 'data_uid', 'nonce', 'enc_uid', 'mic', 'enc_payload',
             'gw_packet', 'stat_param', 'crc_environment_previous', 'temperature_sensor', 'humidity_sensor',
             'light_intensity_sensor']

        self.update_run_data()
        self.update_packets_data(only_titles=True)

    def update_run_data(self, res=None, save_to_csv=True):
        if res is not None:
            test_info = res.get_test_info()
            if not res.is_results_empty():
                self.total_tested += 1
            self.run_data['total_run_passed_offline'] += res.get_total_fail_bin() == FailureCodes.PASS.value
            self.run_data['total_run_bad_printing'] += res.get_total_fail_bin() == FailureCodes.BAD_PRINTING.value
            if 'responded' in test_info:
                self.run_data['total_run_responding_tags'] = test_info['responded']
            if 'missing_label' in test_info:
                self.run_data['total_missing_labels'] = test_info['missing_label']
            if 'ttfp_avg' in test_info:
                self.run_data['ttfp_avg'] = test_info['ttfp_avg']

            self.run_data['total_run_tested'] = self.total_tested - self.run_data['total_missing_labels']
            self.run_data['run_responsive_tags_yield'] = \
                self.run_data['total_run_responding_tags'] / self.run_data['total_run_tested'] * 100 \
                    if self.run_data['total_run_tested'] > 0 else float(-1)
            self.run_data['run_offline_yield'] = \
                self.run_data['total_run_passed_offline'] / self.run_data['total_run_tested'] * 100 \
                    if self.run_data['total_run_tested'] > 0 else float(-1)
            self.run_data['reel_run_end_time'] = datetime.datetime.now()

        if save_to_csv:
            dict_to_csv(dict_in=self.run_data, path=self.run_data_path)

    def update_packets_data(self, res=None, only_titles=False):

        def save_default_packet_data(summary=None):
            default_data = {'raw_packet': None, 'adv_address': None, 'decrypted_packet_type': None,
                            'group_id': None, 'flow_ver': None,
                            'test_mode': None, 'en': None, 'type': None, 'data_uid': None, 'nonce': None,
                            'enc_uid': None, 'mic': None, 'enc_payload': None, 'gw_packet': None, 'rssi': None,
                            'stat_param': None, 'time_from_start': None, 'counter_tag': None,
                            'is_valid_tag_packet': None, 'common_run_name': self.test_data['common_run_name'],
                            'tag_run_location': self.test_data['tag_run_location'],
                            'tag_reel_location': self.test_data['tag_reel_location'],
                            'total_test_duration': self.test_data['total_test_duration'],
                            'status_offline': self.test_data['status_offline'],
                            'fail_bin': self.test_data['fail_bin'], 'fail_bin_str': self.test_data['fail_bin_str'],
                            'external_id': None,
                            'label_validated': self.test_data['label_validated'],
                            'test_num': self.test_data['test_num'],
                            'trigger_time': self.test_data['trigger_time'], 'packet_status': None,
                            'crc_environment_previous': self.test_data.get('crc_environment_previous', None),
                            'temperature_sensor': float('nan'), 'humidity_sensor': float('nan'),
                            'light_intensity_sensor': float('nan')}
            if summary is None:
                default_sum = {'is_test_pass': None, 'selected_tag': None, 'test_start_time': None,
                               'test_end_time': None, 'test_status': None, 'rx_channel': None,
                               'energizing_pattern': None, 'time_profile': None, 'num_packets': 0,
                               'num_cycles': 0, 'sprinkler_counter_mean': None, 'sprinkler_counter_std': None,
                               'sprinkler_counter_min': None, 'sprinkler_counter_max': None, 'tbp_mean': None,
                               'tbp_std': None, 'tbp_min': None, 'tbp_max': None, 'tbp_num_vals': None,
                               'per_mean': None, 'per_std': None, 'rssi_mean': None, 'rssi_std': None,
                               'rssi_min': None, 'rssi_max': None, 'ttfp': None, 'rx_rate_normalized': None,
                               'rx_rate': None}
            else:
                default_sum = summary

            default_packet_att = {'is_valid_packet': None, 'inlay_type': None}
            default_dict = {**default_data, **default_sum, **default_packet_att}
            default_ordered_dict = {k: default_dict[k] for k in self.packet_headers_default}
            dict_to_csv(dict_in=default_ordered_dict, path=self.packets_data_path, append=(not only_titles),
                        only_titles=only_titles)

        if res is not None:
            stat = res.get_total_test_status()
            self.test_data['status_offline'] = int(stat) if stat is not None else 0
            self.test_data['total_test_duration'] = res.get_total_test_duration()
            fail_bin = res.get_total_fail_bin(as_name=True)
            if fail_bin is None:
                fail_bin = FailureCodes.GW_ERROR.name
            self.test_data['fail_bin'] = FailureCodes[fail_bin].value
            self.test_data['fail_bin_str'] = fail_bin
            self.test_data['trigger_time'] = res.get_trigger_time()

            test_info = res.get_test_info()
            self.test_data['tag_run_location'] = test_info['tag_run_location']

            # Include new sensor readings in the test data
            self.test_data['tag_reel_location'] = test_info['tag_reel_location']
            self.test_data['external_id'] = test_info['external_id'] \
                if 'external_id' in test_info.keys() and self.test_data['status_offline'] == 1 else ''
            self.test_data['label_validated'] = test_info['validated'] \
                if 'validated' in test_info.keys() else ValidatedBin.UNTESTED.name
            self.test_data['temperature_sensor'] = test_info['temperature_sensor'] \
                if 'temperature_sensor' in test_info.keys() else float('nan')
            self.test_data['humidity_sensor'] = test_info['humidity_sensor'] \
                if 'humidity_sensor' in test_info.keys() else float('nan')
            self.test_data['light_intensity_sensor'] = test_info['light_sensor'] \
                if 'light_sensor' in test_info.keys() else float('nan')
            self.test_data['crc_environment_previous'] = test_info['crc_environment_previous'] \
                if 'crc_environment_previous' in test_info.keys() else None

            if len(res) == 0:
                save_default_packet_data()

            for i, r in enumerate(res):
                self.test_data['test_num'] = i
                r_sum = r.get_summary()

                # add packets info:
                if r.all_packets.size():
                    # we received packets
                    for p in r.all_packets:
                        p.add_custom_data(
                            custom_data={**self.test_data, **r_sum})
                    r.all_packets.to_csv(path=self.packets_data_path, append=True, export_packet_id=False,
                                         columns=self.packet_headers_default)
                else:
                    # no responds in current test
                    save_default_packet_data(summary=r_sum)

        else:
            self.test_data['test_num'] = 0
            save_default_packet_data()

    def update_default_test_data(self, new_test_data):
        for k, v in new_test_data.items():
            if k not in self.test_data:
                continue
            self.test_data[k] = v

    def exit(self, unsaved_data):
        for tag_result_to_log in unsaved_data:
            if tag_result_to_log is not None:
                # update packet data
                self.update_packets_data(res=tag_result_to_log)
                # update run data
                self.update_run_data(res=tag_result_to_log)



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
import os.path
from threading import Thread
import logging
import threading
from tkinter import *
import serial.tools.list_ports
import pygubu
from os.path import abspath
from json import dump
from os.path import join, dirname
from wiliot_tools.test_equipment.test_equipment import *
from wiliot_core import WiliotGateway, ActionType, DataType, CommandDetails
from wiliot_testers.sample.configs_gui import CONFIGS_DIR

barcodeMutex = threading.Lock()

CONNECT_HW = 'Connect HW'
GO = 'Go'
CONTINUE = 'Continue'
READ = 'Read'
FINISH = 'Finish'

ADD = 'ADD'
REMOVE = 'REMOVE'

GW_CANCEL = '!reset'

# gateway
GW_VERSION = 'Gateway version: '
GW_AVAILABLE_VERSION = 'Available Version: '

# attenuator
ATTENUATION = 'atten'
ATTENUATION_PARAMS = ['attenBle', 'attenLoRa']

BLE = 'Ble'
LORA = 'LoRa'

BARCODES = 'Barcodes'
ATTENUATORS = 'Atten'
CHAMBERS = 'Chambers'

GW_TBP_VERSION = '2.5.1'
TIME_TO_CLOSE_CHAMBER = 3
MAX_BLE_POWER = 22
MAX_LORA_POWER = 29


class ComConnect(object):
    '''
    classdocs
    '''
    isGui = False
    hwConnected = False
    cur_gw_tbp_version = False
    missing_com_port = False
    wiliotTags = False
    gateway = None
    attenuator = None
    chambers_move_com = ''
    barcodes_move_com = ''
    gwVersion = ''
    reel_id = ''
    gtin = ''
    barcodes_state = ADD
    chambers_state = ADD
    barcodes_serials = {}
    chambers_serials = {}
    atten_serials = {}
    sensors_serials = {}
    used_ports = []
    gw_com_port = []
    barcode_error = []
    chambers_to_close = []
    gw_latest_version = ['']
    gw_update_status = 'disabled'
    start_time = 0
    temperature_sensor_readings = []
    all_sensors = []
    ttk = None
    logger_dir = ''

    def __init__(self, top_builder=None, new_tag_func=None, update_go=None, default_dict=None, logger=None,
                 logger_dir=None, tk_frame=None):
        '''
        Constructor
        '''
        self.logger_dir = logger_dir
        self.top_builder = top_builder
        self.add_tag_to_test = new_tag_func
        self.update_go_state = update_go
        self.default_dict = default_dict
        self.tk_frame = tk_frame
        if logger is None:
            self.logger = logging.getLogger('sample')
        else:
            self.logger = logger

    def __del__(self):
        if self.gateway is not None and self.is_gw_serial_open():
            self.gateway.exit_gw_api()
            self.gateway = None

        for com_port, barcode in self.barcodes_serials.items():
            if barcode is not None and barcode.is_open():
                barcode.close_port()
        self.barcodes_serials = {}

        for com_port, chamber in self.chambers_serials.items():
            if chamber is not None and chamber.is_connected():
                chamber.open_chamber()
                chamber.close_port()
        self.chambers_serials = {}

        for com_port, atten in self.atten_serials.items():
            if atten != None and 'serial' in atten.keys() and atten['serial'] != None and atten:
                atten['serial'].GetActiveTE().close_port()
        self.atten_serials = {}

    def gui(self):
        self.builder = builder = pygubu.Builder()
        ui_file = join(abspath(dirname(__file__)), 'utils', 'com_connect.ui')
        self.builder.add_from_file(ui_file)

        img_path = join(abspath(dirname(__file__)), '')
        builder.add_resource_path(img_path)
        img_path = join(abspath(dirname(__file__)), 'utils')
        builder.add_resource_path(img_path)

        if self.tk_frame is not None:
            self.ttk = Toplevel(self.tk_frame)
        else:
            self.ttk = Tk()

        self.ttk.title("ComConnect")

        self.mainwindow = self.builder.get_object('mainwindow', self.ttk)

        self.builder.connect_callbacks(self)
        self.isGui = True
        self.find_com_ports_cb()
        self.set_gui_defaults()

        self.ttk.protocol("WM_DELETE_WINDOW", self.close)
        self.ttk.lift()
        self.ttk.attributes("-topmost", True)
        self.ttk.attributes("-topmost", False)

        # self.set_gui_defaults()

        self.ttk.mainloop()

    def set_gui_defaults(self):

        if self.serials_connected(CHAMBERS):
            self.builder.get_object('connect_chambers').configure(text='Disconnect')
            self.builder.get_object('chosenChambers')['state'] = 'disabled'
            self.builder.get_object('availableChambers')['state'] = 'disabled'
            self.builder.get_object('chambers_up')['state'] = 'disabled'
            self.builder.get_object('chambers_down')['state'] = 'disabled'
            self.builder.get_object('addChambers')['state'] = 'disabled'
        else:
            self.builder.get_object('connect_chambers').configure(text='Connect')

        ble_atten = ''
        lora_atten = ''
        for com_port, atten in self.atten_serials.items():
            ble_atten = com_port if atten['type'] == BLE else ble_atten
            lora_atten = com_port if atten['type'] == LORA else lora_atten

        self.builder.get_object('attenComBle').set(ble_atten)
        self.builder.get_object('attenComLoRa').set(lora_atten)
        if self.serials_connected(ATTENUATORS):
            self.builder.get_object('connect_atten').configure(text='Disconnect')
            self.builder.get_object('attenComBle')['state'] = 'disabled'
            self.builder.get_object('attenComLoRa')['state'] = 'disabled'
        else:
            self.builder.get_object('connect_atten').configure(text='Connect')
        pass

    def update_com_gui(self, available_ports, com_ports):
        self.builder.get_object('gwCom')['values'] = available_ports + ['']
        self.builder.get_object('attenComBle')['values'] = available_ports + ['']
        self.builder.get_object('attenComLoRa')['values'] = available_ports + ['']
        self.update_multi_serials(available_ports, BARCODES)
        self.update_multi_serials(available_ports, CHAMBERS)

        self.missing_com_port = False

        self.check_chosen_ports(com_ports)
        self.check_opened_ports()

    def check_chosen_ports(self, com_ports):
        if len(self.gw_com_port) == 0 or self.gw_com_port[0] not in com_ports:
            self.gw_com_port = ['']
            self.builder.get_object('gwCom').set('')
            self.missing_com_port = True

        i = 0
        while i < len(self.atten_serials.keys()):
            port = list(self.atten_serials.keys())[i]
            atten = list(self.atten_serials.values())[i]
            if port != '' and port not in com_ports:
                self.atten_serials.pop(port)
                self.builder.get_object(f"attenCom{atten['type']}").set('')
                self.missing_com_port = True
                continue
            i += 1

        self.check_multi_coms(BARCODES, com_ports)
        self.check_multi_coms(CHAMBERS, com_ports)

    def check_multi_coms(self, obj, com_ports):
        self.builder.get_object(f'chosen{obj}').delete(0, END)
        ports = getattr(self, f'{obj.lower()}_serials')
        for port in ports.keys():
            if port in com_ports:
                self.builder.get_object(f'chosen{obj}').insert(END, port)
            else:
                self.missing_com_port = True

    def check_opened_ports(self):
        if self.is_gui_opened():
            self.check_gw_open()
            self.check_multi_open(BARCODES)
            self.check_multi_open(CHAMBERS)
            self.check_atten_open()

    def check_multi_open(self, obj):
        if self.serials_connected(obj):
            self.builder.get_object(f'connect_{obj.lower()}').configure(text='Disconnect')
            self.builder.get_object(f'chosen{obj}')['state'] = 'disabled'
            self.builder.get_object(f'available{obj}')['state'] = 'disabled'
            self.builder.get_object(f'{obj.lower()}_up')['state'] = 'disabled'
            self.builder.get_object(f'{obj.lower()}_down')['state'] = 'disabled'
            self.builder.get_object(f'add{obj}')['state'] = 'disabled'
        else:
            self.builder.get_object(f'connect_{obj.lower()}').configure(text='Connect')
            self.builder.get_object(f'chosen{obj}')['state'] = 'normal'
            self.builder.get_object(f'available{obj}')['state'] = 'normal'
            self.builder.get_object(f'{obj.lower()}_up')['state'] = 'normal'
            self.builder.get_object(f'{obj.lower()}_down')['state'] = 'normal'
            self.builder.get_object(f'add{obj}')['state'] = 'normal'

    def check_gw_open(self):
        if len(self.gw_com_port) > 0 and self.gw_com_port[0] != '':
            self.builder.get_object('gwCom').set(self.gw_com_port[0])
        if self.is_gw_serial_open():
            self.builder.get_object('connect_gw').configure(text='Disconnect')
            self.builder.get_object('gwCom')['state'] = 'disabled'
            self.builder.get_object('version').configure(text=GW_VERSION + self.gwVersion[0])
            self.builder.get_object('latestVersion').configure(text=GW_AVAILABLE_VERSION + self.gw_latest_version[0])
            self.builder.get_object('update_gw')['state'] = self.gw_update_status
        else:
            self.builder.get_object('connect_gw').configure(text='Connect')
            self.builder.get_object('gwCom')['state'] = 'normal'
            self.builder.get_object('version').configure(text=GW_VERSION)
            self.builder.get_object('latestVersion').configure(text=GW_AVAILABLE_VERSION)
            self.builder.get_object('update_gw')['state'] = 'disabled'

    def check_atten_open(self):
        connected = False
        for com, atten in self.atten_serials.items():
            if atten['serial'] is not None and atten['serial'].GetActiveTE().is_open():
                connected = True
        if connected:
            self.builder.get_object(f'connect_atten').configure(text='Disconnect')
            self.builder.get_object(f'attenComLoRa')['state'] = 'disabled'
            self.builder.get_object(f'attenComBle')['state'] = 'disabled'
        else:
            self.builder.get_object(f'connect_atten').configure(text='Connect')
            self.builder.get_object(f'attenComLoRa')['state'] = 'normal'
            self.builder.get_object(f'attenComBle')['state'] = 'normal'

    def choose_com_ports(self):
        default_dict = self.default_dict
        com_ports = [comport.device for comport in serial.tools.list_ports.comports()]
        if len(com_ports) == 0:
            com_ports = [s.name for s in serial.tools.list_ports.comports()]
        if 'gw' in default_dict.keys() and default_dict['gw'] in com_ports:
            self.gw_com_port = [default_dict['gw']]
            self.used_ports.append(default_dict['gw'])
        else:
            self.gw_com_port = ['']
            self.missing_com_port = True
        if 'atten' in default_dict.keys() and BLE in default_dict['atten'].keys() and default_dict['atten'][BLE] \
                in com_ports:
            self.atten_serials[default_dict['atten'][BLE]] = {}
            self.atten_serials[default_dict['atten'][BLE]]['type'] = BLE
            self.atten_serials[default_dict['atten'][BLE]]['serial'] = None
            self.used_ports.append(default_dict['atten'][BLE])
        elif 'atten' in default_dict.keys() and BLE in default_dict['atten'].keys() \
                and default_dict['atten'][BLE].strip() != '':
            self.missing_com_port = True
        if 'atten' in default_dict.keys() and LORA in default_dict['atten'].keys() and default_dict['atten'][LORA] \
                in com_ports:
            self.atten_serials[default_dict['atten'][LORA]] = {}
            self.atten_serials[default_dict['atten'][LORA]]['type'] = LORA
            self.atten_serials[default_dict['atten'][LORA]]['serial'] = None
            self.used_ports.append(default_dict['atten'][LORA])
        elif 'atten' in default_dict.keys() and LORA in default_dict['atten'].keys() \
                and default_dict['atten'][LORA].strip() != '':
            self.missing_com_port = True

        if 'barcodes' in default_dict.keys():
            self.barcodes_serials = dict.fromkeys([barcode for barcode in default_dict['barcodes'] if barcode
                                                   in com_ports], None)
            self.used_ports += list(self.barcodes_serials.keys())

        if 'chambers' in default_dict.keys():
            self.chambers_serials = dict.fromkeys([chamber for chamber in default_dict['chambers'] if chamber
                                                   in com_ports], None)
            self.used_ports += list(self.chambers_serials.keys())

        if 'temperature_sensors' in default_dict.keys():
            self.sensors_serials = dict.fromkeys([sensor for sensor in default_dict['temperature_sensors']], None)

        missing_barcodes = []
        missing_chambers = []
        if 'barcodes' in default_dict.keys():
            missing_barcodes = [barcode for barcode in default_dict['barcodes'] if barcode not in com_ports]
        if 'chambers' in default_dict.keys():
            missing_chambers = [chamber for chamber in default_dict['chambers'] if chamber not in com_ports]
        if any(missing_barcodes + missing_chambers):
            self.missing_com_port = True

        return self.missing_com_port

    def init_gw(self):
        return WiliotGateway(logger_name=self.logger.name, mp_reset_time_upon_gw_start=True,
                             np_max_packet_in_buffer_before_error=10,
                             log_dir_for_multi_processes=self.logger_dir)

    def connect_all(self, gui=True):
        if self.gateway is None:
            self.gateway = self.init_gw()
        if not self.is_gw_serial_open():
            success = self.connect_gw(gui)
            if not success:
                return
        if not self.serials_connected(ATTENUATORS):
            self.connect_atten(gui)
        if not self.serials_connected(BARCODES):
            self.connect_barcodes(gui)
        if not self.serials_connected(CHAMBERS):
            self.connect_chambers(gui)
        self.connect_temperature_sensor()
        self.hwConnected = True

    def connect_gw(self, gui=True, disconnect=False):
        if self.gateway is None:
            self.gateway = self.init_gw()
        if not self.is_gw_serial_open() and not disconnect:
            if len(self.gw_com_port) == 0 or self.gw_com_port[0].strip() == '':
                self.popup_message('No default com port for GW, please choose GW com port.', title='Error', log='error')
                return False
            com_port = self.gw_com_port[0]
            self.gateway.open_port(port=com_port, baud=921600)
            if self.is_gw_serial_open():
                self.start_listener(not_print_str=True)
                self.logger.info(f'GW is connected on port: {com_port}.')
                self.gateway.reset_gw()
                sleep(1)
                version = self.gateway.get_gw_version()
                self.gwVersion = version
                self.gw_latest_version = latest_version = self.gateway.get_latest_version_number()
                cur_version = int(version[0].replace('.', ''))
                self.gw_update_status = 'normal' if cur_version < int(latest_version[0].replace('.', '')) \
                    else 'disabled'
                if cur_version >= int(GW_TBP_VERSION.replace('.', '')):
                    self.cur_gw_tbp_version = True
            else:
                self.logger.error(f'Error connecting to GW on port: {com_port}.')
                return False
        else:
            self.gateway.close_port()
            self.builder.get_object('connect_gw').configure(text='Connect')
            self.builder.get_object('version').configure(text=GW_VERSION)
            self.builder.get_object('latestVersion').configure(text=GW_AVAILABLE_VERSION)
            self.builder.get_object('gwCom')['state'] = 'normal'
        if gui:
            self.check_gw_open()

        return True

    def connect_barcodes(self, gui=True):
        self.connect_multi_serials(BARCODES, gui=gui)

    def connect_chambers(self, gui=True):
        self.connect_multi_serials(CHAMBERS, gui=gui)

    def connect_atten(self, gui=True):
        is_connected = self.connect_multi_serials(ATTENUATORS, gui=gui)
        if gui:
            atten_state = 'disabled' if is_connected else 'normal'
            self.builder.get_object('attenComLoRa')['state'] = atten_state
            self.builder.get_object('attenComBle')['state'] = atten_state

    def connect_temperature_sensor(self):
        self.all_sensors = []
        self.temperature_sensor_readings = []
        for sensor_name in self.sensors_serials.keys():
            self.temperature_sensor_readings.append([])
            try:
                sensor_temp = YoctoTemperatureSensor()
                is_temp_sensor_connected = sensor_temp.connect(target=sensor_name)
                if is_temp_sensor_connected:
                    self.logger.info('Temperature Sensor {} is connected'.format(sensor_temp.get_sensor_name()))
                    self.all_sensors.append(sensor_temp)
                else:
                    self.popup_message('Could not connect to Temperature Sensor according to '
                                       'the following name: {}'.format(sensor_name), title='Error', log='error')
                    self.all_sensors.append(None)
                    raise ConnectionError("Could not establish Temperature sensor connection")
            except Exception as e:
                self.logger.info('while connecting to the Temperature Sensor the following error occurs : {}'.format(e))
                self.all_sensors.append(None)
                raise Exception(f'Could not establish Temperature sensor connection due to {e}')

    def read_temperature_sensor(self):
        for i, sensor in enumerate(self.all_sensors):
            if sensor is not None:
                self.temperature_sensor_readings[i].append(sensor.get_temperature())

    def connect_multi_serials(self, obj, gui=True):
        serials = getattr(self, f'{obj.lower()}_serials')
        is_connected = False
        if self.serials_connected(obj):
            self.close_serials(obj, serials)
            self.builder.get_object(f'connect_{obj.lower()}').configure(text='Connect')
            try:
                self.builder.get_object(f'chosen{obj}')['state'] = 'normal'
                self.builder.get_object(f'available{obj}')['state'] = 'normal'
                self.builder.get_object(f'add{obj}')['state'] = 'normal'
            except:
                pass
        elif len(serials.keys()) > 0:
            self.open_serials(obj, serials)
            if self.serials_connected(obj) and gui:
                # if gui:
                is_connected = True
                self.builder.get_object(f'connect_{obj.lower()}').configure(text='Disconnect')
                try:
                    self.builder.get_object(f'chosen{obj}')['state'] = 'disabled'
                    self.builder.get_object(f'available{obj}')['state'] = 'disabled'
                    self.builder.get_object(f'add{obj}')['state'] = 'disabled'
                except:
                    pass
        if gui:
            self.find_com_ports_cb()
        # self.update_go_state()
        return is_connected

    def open_serials(self, obj, serials):
        threads = []
        for com_port, com_serial in serials.items():
            if 'barcode' in obj.lower():
                if com_serial is not None and com_serial.is_open():
                    continue
                com_serial = BarcodeScanner(com_port=com_port, log_type='LOG_NL')
                if com_serial.is_open():
                    # self.used_ports.append(com_port)
                    serials[com_port] = com_serial
            elif 'chamber' in obj.lower():
                if com_serial is not None and com_serial.is_connected():
                    continue
                temp_thread = Thread(target=self.connect_chamber, args=([com_port, serials]))
                temp_thread.start()
                threads.append(temp_thread)
            elif 'atten' in obj.lower():
                if com_serial['serial'] is not None or com_port.strip() == '':
                    # 8 if serial['serial']!=None and
                    # serial['serial'].GetActiveTE().s.is_open():
                    continue
                com_serial = Attenuator('API', comport=com_port)
                if com_serial.GetActiveTE().is_open():
                    # self.used_ports.append(com_port)
                    serials[com_port]['serial'] = com_serial

        for thread in threads:
            thread.join()

    def close(self):
        if self.is_gw_serial_open() and self.serials_connected(ATTENUATORS) and self.serials_connected(BARCODES):
            self.hwConnected = True
            self.enable_hw_connected()
        if self.isGui:
            self.isGui = False
            self.ttk.destroy()
        return self.hwConnected

    def save(self):
        self.default_dict['gw'] = self.gw_com_port[0]
        self.default_dict['atten'] = {}
        for com_port, atten in self.atten_serials.items():
            self.default_dict['atten'][atten['type']] = com_port
        self.default_dict['barcodes'] = list(self.barcodes_serials.keys())
        self.default_dict['chambers'] = list(self.chambers_serials.keys())
        with open(join(CONFIGS_DIR, '.defaults.json'), 'w+') as defaultComs:
            dump(dict(self.default_dict), defaultComs, indent=4)

        self.logger.info(f'Com ports saved successfully.')

    def focus_available(self, obj):
        self.builder.get_object(f'add{obj}').configure(text='>')
        setattr(self, f'{obj.lower()}_state', ADD)

    def focus_chosen(self, obj):
        self.builder.get_object(f'add{obj}').configure(text='<')
        setattr(self, f'{obj.lower()}_state', REMOVE)
        setattr(self, f'{obj.lower()}_move_com', '')

    def add_barcode(self):
        if getattr(self, f'{BARCODES.lower()}_state') == ADD:
            com_chosen = self.builder.get_object(f'available{BARCODES}').get(ACTIVE)
            try:
                temp_barcode = BarcodeScanner(com_port=com_chosen)
            except Exception as e:
                self.popup_message(f'Could NOT connect. {e}', title='Error', log='error')
        self.add(BARCODES)
        self.find_com_ports_cb()

    def add_chamber(self):
        self.add(CHAMBERS)
        self.find_com_ports_cb()

    def add(self, obj):
        if getattr(self, f'{obj.lower()}_state') == ADD:
            sending = self.builder.get_object(f'available{obj}')
            receiving = self.builder.get_object(f'chosen{obj}')
        else:
            sending = self.builder.get_object(f'chosen{obj}')
            receiving = self.builder.get_object(f'available{obj}')

        com_list = list(sending.get(0, END))
        com_chosen = sending.get(ACTIVE)
        receiving.insert(END, com_chosen)
        com_index = com_list.index(com_chosen)
        sending.delete(com_index, com_index)

        serials = getattr(self, f'{obj.lower()}_serials')
        com_ports = self.builder.get_object(f'chosen{obj}').get(0, END)
        old_ports = [port for port in serials.keys() if port not in com_ports]
        new_serials = dict(zip(com_ports, [None] * len(com_ports)))
        self.used_ports = [port for port in self.used_ports if port not in old_ports] + list(com_ports)
        setattr(self, f'{obj.lower()}_serials', new_serials)

    def chamber_up(self):
        self.up(CHAMBERS)

    def chamber_down(self):
        self.down(CHAMBERS)

    def barcode_up(self):
        self.up(BARCODES)

    def barcode_down(self):
        self.down(BARCODES)

    def up(self, obj):
        com_list = list(self.builder.get_object(f'chosen{obj}').get(0, END))
        if getattr(self, f'{obj.lower()}_move_com') == '':
            chosen_com = self.builder.get_object(f'chosen{obj}').get(ACTIVE)
            setattr(self, f'{obj.lower()}_move_com', chosen_com)
        else:
            chosen_com = getattr(self, f'{obj.lower()}_move_com')
        if chosen_com != '':
            com_index = com_list.index(chosen_com)
            if com_index > 0:
                com_list.pop(com_list.index(chosen_com))
                com_list.insert(com_index - 1, chosen_com)
                self.builder.get_object(f'chosen{obj}').delete(0, END)
                for com in com_list:
                    self.builder.get_object(f'chosen{obj}').insert(END, com)
                self.builder.get_object(f'chosen{obj}').select_set(com_index - 1)

    def down(self, obj):
        com_list = list(self.builder.get_object(f'chosen{obj}').get(0, END))
        if getattr(self, f'{obj.lower()}_move_com') == '':
            chosen_com = self.builder.get_object(f'chosen{obj}').get(ACTIVE)
            setattr(self, f'{obj.lower()}_move_com', chosen_com)
        else:
            chosen_com = getattr(self, f'{obj.lower()}_move_com')
        if chosen_com != '':
            com_index = com_list.index(chosen_com)
            if com_index < (len(com_list) - 1):
                com_list.pop(com_list.index(chosen_com))
                com_list.insert(com_index + 1, chosen_com)
                self.builder.get_object(f'chosen{obj}').delete(0, END)
                for com in com_list:
                    self.builder.get_object(f'chosen{obj}').insert(END, com)
                self.builder.get_object(f'chosen{obj}').select_set(com_index + 1)

    def update_multi_serials(self, available_ports, obj):
        self.builder.get_object(f'available{obj}').delete(0, END)
        for port in available_ports:
            if port not in self.used_ports:
                self.builder.get_object(f'available{obj}').insert(END, port)

    def update_atten_serials(self, available_ports):
        for com, item in self.atten_serials.items():
            if item['serial'] is None or not item['serial'].is_open():
                self.builder.get_object(f'attenCom{item["type"]}')['values'] = available_ports
            else:
                self.builder.get_object(f'attenCom{item["type"]}').set(com)

    def connect_chamber(self, com_port, serials):
        com_serial = Tescom(com_port)
        if com_serial.is_connected():
            # self.used_ports.append(com_port)
            serials[com_port] = com_serial
            if not com_serial.is_door_open():
                com_serial.open_chamber()

    def close_serials(self, obj, serials):
        if 'atten' in obj.lower():
            for serial in serials.values():
                # serial['serial'].GetActiveTE.s.close_port()
                if serial['serial'] is not None and serial['serial'].GetActiveTE().is_open():
                    serial['serial'].GetActiveTE().close_port()
                serial['serial'] = None
                # self.used_ports.remove(com_port)
        else:
            for com_serial in serials.values():
                if 'chamber' in obj.lower():
                    com_serial.open_chamber()
                com_serial.close_port()
                # self.used_ports.remove(com_port)
                # serials.pop(com_port)

    def serials_connected(self, obj):
        serials = getattr(self, f'{obj.lower()}_serials')
        if 'atten' in obj.lower():
            serials = dict(zip(serials.keys(), [atten['serial'] for atten in serials.values()]))
        connected_serials = 0
        for com_port, com_serial in serials.items():
            if com_serial is not None:
                if 'barcode' in obj.lower() and com_serial.is_open():
                    connected_serials += 1
                if 'chamber' in obj.lower() and com_serial.is_connected():
                    connected_serials += 1
                if 'atten' in obj.lower():
                    connected_serials += 1
            # if com_port.strip()=='':
            #     connected_serials += 1
        if connected_serials > 0 and connected_serials == len(serials.keys()):
            return True
        else:
            return False

    def read_barcode(self, scanner_index=0, close_chamber=False, add_to_test=False, n_try=5):
        full_data, reel_id, gtin = None, None, None
        if len(list(self.barcodes_serials.values())) == 0:
            self.logger.error('Trying to read barcode but no scanner were connected')
            return None, None
        scanner = list(self.barcodes_serials.values())[scanner_index]
        for i in range(n_try):
            full_data, cur_id, reel_id, gtin = scanner.scan_ext_id()
            if full_data is None and cur_id is None and reel_id is None and full_data is None:
                continue
            break

        reel_id = reel_id if reel_id is not None else full_data
        gtin = gtin if gtin is not None else ''
        if reel_id is not None:
            self.reel_id = reel_id
            self.gtin = gtin
        if full_data is None:
            barcodeMutex.acquire()
            if close_chamber:
                self.barcode_error.append(scanner_index)
            barcodeMutex.release()
            return None, None

        if not close_chamber:
            reel_id_obj = self.top_builder.tkvariables.get('reelId')
            reel_id_obj.set(self.reel_id)

        else:
            success = self.add_tag_to_test(full_data, reel_id, scanner_index=scanner_index, add_to_test=add_to_test)
            if success:
                self.chambers_to_close.append(scanner_index)

        return full_data, gtin + reel_id

    def get_all_scanners_index(self):
        return list(range(len(self.barcodes_serials.values())))

    def read_scanners_barcodes(self, indexes=()):
        if len(indexes) == 0:
            indexes = self.get_all_scanners_index()
        scanner_threads = []
        self.barcode_error = []
        self.chambers_to_close = []

        for i in indexes:
            t = threading.Thread(target=self.read_barcode, args=(i, True, True))
            scanner_threads.append(t)
            t.start()
        for i in range(len(scanner_threads)):
            t = scanner_threads[i]
            t.join()

        read_message = ''
        title = 'Warning'
        font = 18
        if len(self.barcode_error) > 0:
            read_message += f'Error reading external ID from chambers {self.barcode_error},' \
                            f' try repositioning the tags.\n'
            title = 'Error'
            font = 16

        if len(self.chambers_to_close) > 0:
            read_message += f'Chambers are closing!!\nWatch your hands!!!'

        self.popup_message(read_message, title, ("Helvetica", font), title.lower())

        if len(self.chambers_to_close) > 0:
            self.close_chambers(self.chambers_to_close)
        sleep(TIME_TO_CLOSE_CHAMBER)
        self.update_go_state()

    def enable_hw_connected(self):
        self.top_builder.get_object('read_qr')['state'] = 'normal'
        self.top_builder.get_object('reelId')['state'] = 'normal'
        if self.top_builder.tkvariables.get('go').get() == CONNECT_HW:
            self.top_builder.tkvariables.get('go').set(READ)
            self.top_builder.get_object('go')['state'] = 'disabled'

    def set_attenuation(self, params):
        for com_port, atten in self.atten_serials.items():  # BLE and/or Lora
            try:
                value = params[atten['type'].lower() + 'Attenuation']
                attenuation = atten['serial'].GetActiveTE().Setattn(float(value))
                self.logger.info(f"{atten['type']} Attenuation set to: {str(attenuation).strip()}")
            except Exception as e:
                err = f"{atten['type']} Attenuator error: try reconnect the attenuator [{e}].\n" \
                      f"or params are not correct: {params.keys()}: [{e}]"
                self.logger.error(err)
                raise Exception(err)

    def send_gw_app(self, params):
        str_rsp = []
        self.gateway.reset_buffer()
        self.gateway.clear_rsp_str_input_q()

        self.gateway.reset_start_time()

        for param, value in params.items():
            if param.startswith(ATTENUATION):
                for com_port, atten in self.atten_serials.items():
                    if param.lower().endswith(atten['type'].lower()):
                        try:
                            attenuation = atten['serial'].GetActiveTE().Setattn(float(value))
                            self.logger.info(f"{atten['type']} Attenuation set to: {str(attenuation).strip()}")
                        except Exception as e:
                            self.logger.error(f"{atten['type']} Attenuator error: try reconnect the attenuator [{e}].")
                            return False

        self.start_time = time()
        try:
            if 'EmulateSurfaceValue' in params.keys():
                str_rsp.append(self.gateway.write('!set_sub_1_ghz_energizing_frequency {}'.format(
                    params['EmulateSurfaceValue']), with_ack=True))
            cmds = {
                CommandDetails.set_sub_1_ghz_power: [MAX_LORA_POWER],
                CommandDetails.set_energizing_pattern: [params['pattern']],
                CommandDetails.scan_ch: [params['channel']],
                CommandDetails.time_profile: [params['tTotal'], params['tOn']]
            }
            if params.get('symbol', None):
                cmds[CommandDetails.set_scan_radio] = self.gateway.get_cmd_symbol_params(freq_str=params['symbol'])
            output_power_cmds = self.gateway.get_cmds_for_abs_output_power(abs_output_power=MAX_BLE_POWER)
            cmds = {**cmds, **output_power_cmds}
            gw_ans = self.gateway.set_configuration(cmds=cmds, start_gw_app=True)

            for rsp in gw_ans:
                str_rsp.append(rsp)

        except Exception as e:
            self.logger.error(f'Gateway error: try reconnect the gateway [{e}].')
            self.cancel_gw_commands()
            return False

        return True

    def connect_and_close(self):
        self.connect_all()
        self.close()

    def get_reel_id(self):
        return self.reel_id

    def get_gtin(self):
        return self.gtin

    def get_reel_external(self):
        return self.gtin + self.reel_id

    def is_gw_serial_open(self):
        if self.gateway is None:
            return False
        serial_open, _, _ = self.gateway.get_connection_status()
        return serial_open

    def is_gw_data_available(self):
        return self.gateway.is_data_available()

    def is_gui_opened(self):
        return self.isGui

    def get_gw_version(self):
        return self.gwVersion[0]

    def update_gw(self):
        self.gateway.update_version()

    def get_gw_time(self):
        return self.start_time

    def cancel_gw_commands(self):
        # self.gateway.write(GW_CANCEL)
        self.gateway.reset_gw()
        # self.gateway.stop_continuous_listener()
        # self.gateway.reset_buffer()
        sleep(0.1)

    def start_listener(self, not_print_str=True):
        # self.gateway.start_continuous_listener(not_print_str)
        self.gateway.start_continuous_listener()

    def is_hw_connected(self):
        return self.hwConnected

    def get_chambers(self):
        return list(self.chambers_serials.values())

    def open_chambers(self, indexes=()):
        chambers_threads = []
        chambers = list(self.chambers_serials.values())
        if len(indexes) == 0:
            indexes = list(range(len(chambers)))
        for index in indexes:
            if len(chambers) > index and chambers[index] is not None:
                temp_thread = Thread(target=chambers[index].open_chamber, args=())
                temp_thread.start()
                chambers_threads.append(temp_thread)

        for thread in chambers_threads:
            thread.join()

    def close_chambers(self, indexes=()):
        chambers = list(self.chambers_serials.values())
        chambers_threads = []
        if len(indexes) == 0:
            indexes = list(range(len(chambers)))
        for index in indexes:
            if len(chambers) > index and chambers[index] is not None:
                t = threading.Thread(target=chambers[index].close_chamber, args=())
                chambers_threads.append(t)
                t.start()
        for thread in chambers_threads:
            thread.join()

    def get_num_of_barcode_scanners(self):
        return len(self.barcodes_serials.keys())

    def get_error_barcode(self):
        return self.barcode_error

    def gw_tbp_version(self):
        return self.cur_gw_tbp_version

    def gw_get_packets(self):
        self.gateway.get_packets(action_type=ActionType.ALL_SAMPLE, data_type=DataType.PROCESSED)

    def get_default_dict(self):
        return self.default_dict

    def popup_message(self, msg, title='Error', font=("Helvetica", 10), log='info', bg=None, tk_frame=None):
        if tk_frame:
            popup = Toplevel(tk_frame)
        else:
            popup = Tk()
        popup.eval('tk::PlaceWindow . center')
        popup.wm_title(title)
        if bg is not None:
            popup.configure(bg=bg)
        getattr(self.logger, log)(f'{title} - {msg}')

        def popup_exit():
            popup.destroy()

        label = Label(popup, text=msg, font=font)
        label.pack(side="top", fill="x", padx=10, pady=10)
        b1 = Button(popup, text="Okay", command=popup_exit)
        b1.pack(padx=10, pady=10)
        popup.mainloop()

    # ############## GUI Callbacks  #######################

    def find_com_ports_cb(self, *args):
        com_ports = serial_ports()
        available_ports = [com_port for com_port in com_ports if com_port not in self.used_ports]

        self.update_com_gui(available_ports, com_ports)

    def focus_barcode_available_cb(self, *args):
        self.focus_available(BARCODES)

    def focus_barcode_chosen_cb(self, *args):
        self.focus_chosen(BARCODES)

    def focus_chamber_available_cb(self, *args):
        self.focus_available(CHAMBERS)

    def focus_chamber_chosen_cb(self, *args):
        self.focus_chosen(CHAMBERS)

    def choose_gw_cb(self, *args):
        if len(self.gw_com_port) > 0 and self.gw_com_port[0] != '':
            self.used_ports.pop(self.used_ports.index(self.gw_com_port[0]))
        self.gw_com_port = [self.builder.get_object('gwCom').get()]
        self.used_ports.append(self.gw_com_port[0])

    def choose_ble_atten_cb(self, *args):
        ble_com = self.builder.get_object('attenComBle').get()
        ble_last_com = [com for com, item in self.atten_serials.items() if item['type'] == BLE]
        if len(ble_last_com) > 0:
            self.atten_serials.pop(ble_last_com[0])
            self.used_ports.pop(self.used_ports.index(ble_last_com[0]))
        if ble_com.strip() != '':
            self.atten_serials[ble_com] = {}
            self.atten_serials[ble_com]['type'] = BLE
            self.atten_serials[ble_com]['serial'] = None
            self.used_ports.append(ble_com)

    def choose_lora_atten_cb(self, *args):
        lora_com = self.builder.get_object('attenComLoRa').get()
        lora_last_com = [com for com, item in self.atten_serials.items() if item['type'] == LORA]
        if len(lora_last_com) > 0:
            self.atten_serials.pop(lora_last_com[0])
            self.used_ports.pop(self.used_ports.index(lora_last_com[0]))
        if lora_com.strip() != '':
            self.atten_serials[lora_com] = {}
            self.atten_serials[lora_com]['type'] = LORA
            self.atten_serials[lora_com]['serial'] = None
            self.used_ports.append(lora_com)



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
import json
import os
import pathlib
import time
from datetime import datetime
from tkinter import messagebox
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from wiliot_core import WiliotGateway, set_logger, PacketList, DataType, ActionType
from wiliot_tools.utils.wiliot_gui.wiliot_gui import *

from wiliot_tools.test_equipment.test_equipment import EquipmentError, Attenuator
from wiliot_testers.wiliot_tester_log import WiliotTesterLog

# global vars go here :)
global LORA_ATTN_COM
global BLE_ATTN_COM
global STABILIZATION_RANGE_MODIFIER  # Max differance between points that fit the criteria
global MIN_CONSECUTIVE_POINTS_MODIFIER  # How many points in a row should be in order to recommend a value
global OPTIMAL_ATTENUATION_MODIFIER  # By how much attenuation to increase after finding stabilisation point

LORA_ATTN_COM = "COM3"
BLE_ATTN_COM = "COM6"
STABILIZATION_RANGE_MODIFIER = 1
MIN_CONSECUTIVE_POINTS_MODIFIER = 5
OPTIMAL_ATTENUATION_MODIFIER = 2

run_default_values = {
    'time_profile_on': '5',
    'time_profile_period': '15',
    'lora_value': '20',
    'iter_length': '90',
    'energy_pattern_value': '50',
    'ble_sweep_start': 0,
    'ble_sweep_end': 30,
}

multiprocessing = True
now = datetime.now()
run_start_time = now.strftime("%m-%d-%Y_%H-%M-%S")


class BeaconCalibration:

    def __init__(self):
        # Initialises logger and GUI
        self.init_name_path()
        self.init_gui()
        self.gw_logger = WiliotTesterLog(run_name="beacon_calibration_testing")
        self.gw_logger.set_logger(self.gw_log_path, tester_name="Wiliot tester")

        # Initialises the hardware
        self.attn_enable = True
        self.gw_obj = None
        self.attn_obj_lora = None
        self.attn_obj_ble = None
        self.init_hw()
        self.all_results = pd.DataFrame({'Adva': [], 'Frequency': [], 'TBP': [], 'Atten': []})
        self.gw_config()
        self.run_sweep()
        self.process_result()
        self.analysis_plot()

    def init_name_path(self):
        self.results_path = os.path.join('', 'Beacon_calibration_runs',
                                         ('Beacon_calibration_run_{}'.format(run_start_time)))
        if not os.path.exists(self.results_path):
            pathlib.Path(self.results_path).mkdir(parents=True, exist_ok=True)

        self.gw_log_path = os.path.join('', 'Beacon_calibration_logs')  # Gateway log location

        logger_path, self.logger = set_logger(app_name='BeaconCalibration', dir_name='Beacon_calibration_logs',
                                              file_name='Beacon_calibration_test_log')
        self.logger.critical("I love you, unassuming user")

    def init_hw(self):

        self.gw_obj = WiliotGateway(auto_connect=True, logger_name=self.gw_logger.gw_logger.name, verbose=False)
        try:
            time.sleep(1)
            if self.gw_obj.connected:
                self.gw_obj.reset_gw()
                self.gw_obj.reset_buffer()
                self.gw_obj.is_gw_alive()
                self.gw_obj.start_continuous_listener()

                self.gw_logger.gw_logger.info('GW config finished')
            else:
                raise EquipmentError('Gateway Error - Verify WiliotGateway connection')
        except Exception as e:
            self.gw_logger.gw_logger.warning(e)
            self.gw_obj.close_port()
            raise EquipmentError('Gateway Error - Verify WiliotGateway connection')

        if self.attn_enable:
            self.attn_obj_lora = Attenuator(ATTN_type='API', comport=LORA_ATTN_COM).GetActiveTE()
            attn = self.lora_value
            value = self.attn_obj_lora.Setattn(attn=attn)
            if value != attn:
                raise Exception(f'Error setting attenuation: new : {attn} current: {value}')

    def gw_config(self):
        self.gw_obj.config_gw(energy_pattern_val=self.energy_pattern_val,
                              time_profile_val=[self.time_profile_on_val, self.time_profile_period_val])

    def run_sweep(self):
        self.attn_obj_ble = Attenuator(ATTN_type="API", comport=BLE_ATTN_COM).GetActiveTE()
        self.gw_obj.start_continuous_listener()
        sweep_attns = range(self.ble_sweep_start, self.ble_sweep_end, 1)
        self.attenuation_list = []
        self.all_results = PacketList()
        for attn in sweep_attns:
            # set atten
            self.attn_val = self.attn_obj_ble.Getattn()
            self.attn_val = int(self.attn_obj_ble.Setattn(attn=attn))
            self.attenuation_list.append(attn)
            self.logger.info(f'Current setting attenuation: new : {attn} current: {self.attn_val}')

            self.gw_obj.reset_listener()
            self.gw_obj.config_gw(start_gw_app=True)
            self.daq()
            is_stopped = self.gw_obj.stop_gw_app()
            if not is_stopped:
                raise Exception('could not send stop to GW')
        return self.attenuation_list, self.attn_val

    def daq(self):
        time.sleep(2)  # possibly useless, check
        start_time = time.time()  # Records starting time
        max_duration = self.iter_duration
        while True:
            time.sleep(0)

            if self.gw_obj.is_data_available():
                cur_packet_list = self.gw_obj.get_packets(action_type=ActionType.ALL_SAMPLE,
                                                          data_type=DataType.PACKET_LIST)

                if len(cur_packet_list) > 0:
                    for p in cur_packet_list:
                        p.add_custom_data({'attenuation': self.attn_val})
                        self.logger.info("Would you look at that! , i found this packet: {}"
                                         .format(p.packet_data['raw_packet']))
                        # self.logger.info('got packet {}'.format(p.packet_data['raw_packet']))
                        self.all_results.append(p)
            if time.time() - start_time > max_duration:
                break
        self.logger.info('Iteration ended!')

    def process_result(self):
        self.tbp_df = pd.DataFrame()  # creates an empty dataframe

        df = self.all_results.get_df(add_sprinkler_info=True)  # makes a dataframe from the data so far
        df_packet = df.drop_duplicates(subset=['raw_packet'])  # drops duplicates of packets
        self.group_tbp = df_packet.groupby(by=['attenuation', 'adv_address']).agg(
            tbp_mean=('tbp', lambda x: np.ceil(np.mean(x[x > 0]))),  # Calculates mean, then applies ceiling
            tbp_min=('tbp', lambda x: np.ceil(np.min(x[x > 0]))),
            tbp_median=('tbp', lambda x: np.ceil(np.median(x[x > 0])))  # Calculates min, then applies ceiling
        ).reset_index()

        self.group_tbp = self.group_tbp.sort_values(by=['adv_address', 'attenuation'])

        self.group_tbp['Difference'] = self.group_tbp.groupby('adv_address')['tbp_min'].diff().fillna(0)
        self.group_tbp = self.group_tbp.dropna()

        # Various calculations on the TBP packets
        self.group_tbp.to_csv(os.path.join(self.results_path, "tbp_calculated_{}.csv".format(run_start_time)))
        # Saves packets with TBP only
        df_packet.to_csv(os.path.join(self.results_path, "filtered_tbp_packets_{}.csv".format(run_start_time)))
        # Saves all the results
        df.to_csv(os.path.join(self.results_path, "full_Test_results_{}.csv".format(run_start_time)))

    def analysis_plot(self):
        self.optimal_value_calculation()
        for adva in self.group_tbp['adv_address'].unique():
            group_data = self.group_tbp[self.group_tbp['adv_address'] == adva]
            y1 = self.group_tbp['tbp_min']
            y2 = self.group_tbp['tbp_mean']
            y3 = self.group_tbp['tbp_median']
            plt.plot(group_data['attenuation'], y1, marker='o', label="tbp_min")

            # Used for debugging and comparing results
            # plt.plot(group_data['attenuation'], y2, marker='o', label="tbp_mean")
            # plt.plot(group_data['attenuation'], y3, marker='o', label="tbp_median")

        plt.draw()
        plt.xlabel("Attenuation")
        plt.ylabel('TBP')
        if self.recommended_stabilization:
            plt.title(f"Recommended attenuation: {self.recommended_stabilization}")
        else:
            plt.title("No recommended attenuation based on criteria")
        plt.legend()
        # TODO: add recommended attenuation to graph
        plt.savefig((os.path.join(self.results_path, "beacon_calibration_plot_{}.png".format(run_start_time))))
        plt.show()

    def optimal_value_calculation(self):
        self.group_tbp['Incremental_Difference'] = self.group_tbp['tbp_min'].diff()

        stabilization_range = 1  # Max differance between points that fit the criteria
        min_consecutive_points = 5  # How many points in a row should be in order to recommend a value

        streak_count = 0  # Tracks the longest streak of values within the stabilization range
        stabilization_start_index = None
        try:
            for index, row in self.group_tbp.iterrows():
                if abs(row['Incremental_Difference']) <= stabilization_range:
                    streak_count += 1
                    if streak_count == min_consecutive_points:
                        stabilization_start_index = index - min_consecutive_points + 1
                        stabilization_start_value = self.group_tbp.loc[stabilization_start_index, 'attenuation']
                        break
                else:
                    streak_count = 0  # Resets said counter if conditions are not met

        except KeyError:
            pass

        if stabilization_start_index is not None:
            self.logger.critical(
                f"Stabilization starts at Attenuation: {stabilization_start_index}, tbp_min: {int(self.group_tbp.loc[stabilization_start_index, 'tbp_min'])}")
            self.recommended_stabilization = stabilization_start_value + OPTIMAL_ATTENUATION_MODIFIER
            messagebox.showinfo(message=f"Recommended Attenuation: {int(stabilization_start_value) + 2}")

        else:
            print("No stabilization detected based on the criteria :(")

    # GUI
    def init_gui(self):

        json_params_path = os.path.join(os.path.dirname(__file__), 'beacon_power_calibration_params.json')

        def load_params_from_json(json_params_path):
            if os.path.exists(json_params_path):
                try:
                    with open(json_params_path, 'r') as json_file:
                        return json.load(json_file)
                except json.JSONDecodeError:
                    print("Error decoding JSON from file")
            return {}

        used_values = load_params_from_json(json_params_path)

        # TODO
        # add the attenuation ports to the GUI

        params_dict = {
            'time_profile_on': {
                'text': 'On',
                'value': used_values.get('time_profile_on', '5'),
                'widget_type': 'entry',
                'group': 'Time profile'
            },
            'time_profile_period': {
                'text': 'Period',
                'value': used_values.get('time_profile_period', '15'),
                'widget_type': 'entry',
                'group': 'Time profile'
            },
            'lora_value': {
                'text': 'LoRa Value',
                'value': used_values.get('lora_value', '20'),
                'widget_type': 'entry',
                'options': {'font': ('Helvetica', 10, 'bold')},
            },
            'iter_length': {
                'text': 'Iteration length (in seconds)',
                'value': used_values.get('iter_length', '90'),
                'widget_type': 'entry',
                'options': {'font': ('Helvetica', 10, 'bold')},
            },
            'energy_pattern_value': {
                'text': 'Energy pattern',
                'value': used_values.get('energy_pattern_value', '50'),
                'widget_type': 'entry',
                'options': {'font': ('Helvetica', 10, 'bold')},
            },
            'sweep_row': [
                {'ble_sweep_start': {
                    'text': 'BLE Sweep Start',
                    'value': '0',
                    'widget_type': 'spinbox',
                    'options': list(range(0, 31)),
                },
                },
                {'ble_sweep_end': {
                    'text': 'to',
                    'value': '30',
                    'widget_type': 'spinbox',
                    'options': list(range(0, 31)),
                },
                }
            ],

            'LORA_ATTN_COM': {
                'text': 'LoRa COM Port',
                'value': used_values.get('LORA_ATTN_COM', 'COM3'),
                'widget_type': 'entry',
                'options': {'font': ('Helvetica', 10, 'bold')},
                'group': 'COM Port Configuration'
            },
            'BLE_ATTN_COM': {
                'text': 'BLE COM Port',
                'value': used_values.get('BLE_ATTN_COM', 'COM6'),
                'widget_type': 'entry',
                'options': {'font': ('Helvetica', 10, 'bold')},
                'group': 'COM Port Configuration'
            },
        }

        beacon_power_calib_gui = WiliotGui(params_dict=params_dict)
        values = beacon_power_calib_gui.run()
        print(values)
        try:
            form_values = {
                'iter_duration': int(values['iter_length']),
                'lora_value': int(values['lora_value']),
                'sweep_row_ble_sweep_start': values['sweep_row_ble_sweep_start'],
                'sweep_row_ble_sweep_end': values['sweep_row_ble_sweep_end'],
                'energy_pattern_value': int(values['energy_pattern_value']),
                'time_profile_on': int(values['time_profile_on']),
                'time_profile_period': int(values['time_profile_period']),
                'LORA_ATTN_COM': values['LORA_ATTN_COM'],
                'BLE_ATTN_COM': values['BLE_ATTN_COM'],
            }

            # Serialize form values to JSON and write to a file
            with open('beacon_power_calibration_params.json', 'w') as json_file:
                json.dump(form_values, json_file, indent=4)

        except ValueError:
            popup_message('Value must be an integer.', logger=self.logger)


if __name__ == '__main__':
    inst = BeaconCalibration()
    SystemExit("We're done here, get on with your day mate")



import datetime
import time
from collections import deque

import serial
from wiliot_testers.yield_tester.simulation.yield_simulation_utils import AUTO_TRIGGERS, TIME_BETWEEN_AUTO_TRIGGERS
from wiliot_testers.yield_tester.utils.get_arduino_ports import get_arduino_ports

BAUD_ARDUINO = 1000000


class CountThread(object):
    """
    Counting the number of tags
    """

    def __init__(self, stop_event, logger_file, matrix_size=1, ther_cols=1):
        self.arduino_connection_error = False
        self.pause_triggers = False
        self.logger_file = logger_file
        self.start_time = datetime.datetime.now()
        self.last_arduino_trigger_time = datetime.datetime.now()
        self.comPortObj = None
        self.trigger_port = None
        if not AUTO_TRIGGERS:
            self.connect()
        self.matrix_size = matrix_size
        self.ther_cols = ther_cols
        self.stop = stop_event
        self.matrix_cnt = 0
        self.trigger_times = deque(maxlen=10)

    def connect(self):
        optional_ports = get_arduino_ports()
        if len(optional_ports) == 0:
            raise Exception("NO ARDUINO")
        for port in optional_ports:
            try:
                self.comPortObj = serial.Serial(port, BAUD_ARDUINO, timeout=0.1)
                time.sleep(2)
                initial_message = self.comPortObj.readline().decode().strip()
                if "Wiliot Yield Counter" in initial_message:
                    self.trigger_port = port
            except Exception as e:
                raise Exception(f'could not connect to port {port} due to {e}')

    def raising_trigger(self):
        self.last_arduino_trigger_time = datetime.datetime.now()
        self.matrix_cnt += 1
        self.update_trigger_times((self.last_arduino_trigger_time - self.start_time).total_seconds())  # a float
        self.logger_file.info(f'Got a Trigger.  Number of Triggers {self.matrix_cnt}')

    def reconnect(self):
        """
        Attempts to reconnect to the Arduino.
        """
        connected = False
        start_time = time.time()
        while not connected and not self.stop.is_set() and time.time() - start_time < 60:
            try:
                self.comPortObj = serial.Serial(self.trigger_port, BAUD_ARDUINO, timeout=0.1)
                connected = True
                self.logger_file.info("Reconnected to Arduino")
            except serial.SerialException:
                self.logger_file.error("Reconnection failed. Trying again...")
                time.sleep(5)
        if not connected:
            self.arduino_connection_error = True

    def run(self):
        """
        Tries to read data and then counts the number of tags
        """
        while not self.stop.is_set():
            time.sleep(0.100)
            if not AUTO_TRIGGERS:
                try:
                    data = self.comPortObj.readline()
                    if data.__len__() > 0:
                        try:
                            tmp = data.decode().strip(' \t\n\r')
                            if "pulses detected" in tmp and not self.pause_triggers:
                                self.raising_trigger()
                        except Exception as ee:
                            self.logger_file.error(f'Warning: Could not decode counter data or Warning: {ee}')
                except serial.SerialException as e:
                    self.logger_file.error("Arduino is disconnected   ", e)
                    self.reconnect()
                except Exception as ee:
                    self.logger_file.error(f"NO READLINE: {ee}")
            else:
                self.raising_trigger()
                time.sleep(TIME_BETWEEN_AUTO_TRIGGERS)
        if not AUTO_TRIGGERS:
            self.comPortObj.close()

    def update_trigger_times(self, trigger_time):
        self.trigger_times.append([trigger_time, self.matrix_cnt])

    def get_trigger_times(self):
        return self.trigger_times

    def pop_from_triggers_queue(self):
        self.trigger_times.popleft()

    def set_pause_triggers(self, paused):
        self.pause_triggers = paused

    def get_tested(self):
        """
        returns the number of tags
        """
        return self.matrix_cnt * int(self.matrix_size)

    def get_arduino_connection_error(self):
        return self.arduino_connection_error

    def get_matrix_cnt(self):
        return self.matrix_cnt
